import pytest

from schematics.common import *
from schematics.models import Model
from schematics.transforms import blacklist, export_loop, whitelist, wholelist
from schematics.types import IntType, LongType, MD5Type, StringType
from schematics.types.compound import DictType, ListType, ModelType
from schematics.types.serializable import serializable


def test_serializable():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    location_US = Location({"country_code": "US"})

    assert location_US.country_name == "United States"

    d = location_US.serialize()
    assert d == {"country_code": "US", "country_name": "United States"}

    d = location_US.to_native()
    assert d == {"country_code": "US", "country_name": "United States"}

    location_IS = Location({"country_code": "IS"})

    assert location_IS.country_name == "Unknown"

    d = location_IS.serialize()
    assert d == {"country_code": "IS", "country_name": "Unknown"}

    d = location_IS.to_native()
    assert d == {"country_code": "IS", "country_name": "Unknown"}


def test_serializable_to_native():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    loc = Location({"country_code": "US"})

    d = loc.to_native()
    assert d == {"country_code": "US", "country_name": "United States"}


def test_serializable_with_serializable_name():
    class Location(Model):
        country_code = StringType(serialized_name="cc")

        @serializable(serialized_name="cn")
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    location_US = Location({"cc": "US"})

    assert location_US.country_name == "United States"

    d = location_US.serialize()
    assert d == {"cc": "US", "cn": "United States"}


class PlayerIdType(LongType):
    def to_primitive(self, value, context=None):
        return str(value)


def test_serializable_with_custom_serializable_class():
    class Player(Model):
        id = LongType()

        @serializable(type=PlayerIdType())
        def player_id(self):
            return self.id

    player = Player({"id": 1})

    assert player.id == 1
    assert player.player_id == 1

    d = player.serialize()
    assert d == {"id": 1, "player_id": "1"}


def test_serializable_with_type_as_positional_argument():
    class Player(Model):
        id = LongType()

        @serializable(PlayerIdType)
        def player_id(self):
            return self.id

    player = Player({"id": 1})

    assert player.id == 1
    assert player.player_id == 1

    d = player.serialize()
    assert d == {"id": 1, "player_id": "1"}


def test_serializable_with_type_and_options():
    class Player(Model):
        id = LongType()

        @serializable(PlayerIdType(), serialized_name="playerId")
        def player_id(self):
            return self.id

    player = Player({"id": 1})

    assert player.id == 1
    assert player.player_id == 1

    d = player.serialize()
    assert d == {"id": 1, "playerId": "1"}


def test_serializable_with_model():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel(dict(level=self.total_points * 2, title="Best"))

    player = Player({"total_points": 2})

    assert player.xp_level.level == 4

    d = player.serialize()
    assert d == {"total_points": 2, "xp_level": {"level": 4, "title": "Best"}}


def test_serializable_with_model_to_native():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel(dict(level=self.total_points * 2, title="Best"))

    player = Player({"total_points": 2})

    assert player.xp_level.level == 4

    d = player.to_native()
    assert d == {"total_points": 2, "xp_level": {"level": 4, "title": "Best"}}


def test_serializable_with_model_when_None():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return None if not self.total_points else ExperienceLevel()

    player = Player({"total_points": 0})

    assert player.xp_level is None

    d = player.serialize()
    assert d == {"total_points": 0, "xp_level": None}


def test_serializable_with_model_hide_None():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

    class Player(Model):
        total_points = IntType()

        @serializable(type=ModelType(ExperienceLevel), serialize_when_none=False)
        def xp_level(self):
            return None if not self.total_points else ExperienceLevel()

    player = Player({"total_points": 0})

    assert player.xp_level is None

    d = player.serialize()
    assert d == {"total_points": 0}


def test_serializable_with_embedded_models_and_list():
    class Question(Model):
        id = LongType()

    class QuestionPack(Model):
        id = LongType()
        questions = ListType(ModelType(Question))

    class Game(Model):
        id = StringType()
        question_pack = ModelType(QuestionPack)

    q1 = Question({"id": 1})
    q2 = Question({"id": 2})

    game = Game({"id": "1", "question_pack": {"id": 2, "questions": [q1, q2]}})

    assert game.question_pack.questions[0] == q1
    assert game.question_pack.questions[1] == q2

    d = game.serialize()

    assert d == {
        "id": "1",
        "question_pack": {
            "id": 2,
            "questions": [
                {
                    "id": 1,
                },
                {
                    "id": 2,
                },
            ],
        },
    }


def test_serializable_with_embedded_models():
    class ExperienceLevel(Model):
        level = IntType()
        stars = IntType()

        @classmethod
        def from_total_points(cls, total_points, category_slug):
            return cls(dict(level=total_points * 2, stars=total_points))

    class CategoryStatsInfo(Model):
        category_slug = StringType()
        total_points = IntType(default=0)

        @serializable(type=ModelType(ExperienceLevel))
        def xp_level(self):
            return ExperienceLevel.from_total_points(
                self.total_points, self.category_slug
            )

    class PlayerInfo(Model):
        id = LongType()
        display_name = StringType()

    class PlayerCategoryInfo(PlayerInfo):
        categories = DictType(ModelType(CategoryStatsInfo))

    info = PlayerCategoryInfo(
        dict(
            id="1",
            display_name="John Doe",
            categories={"math": {"category_slug": "math", "total_points": 1}},
        )
    )

    assert info.categories["math"].xp_level.level == 2
    assert info.categories["math"].xp_level.stars == 1

    d = info.serialize()
    assert d == {
        "id": 1,
        "display_name": "John Doe",
        "categories": {
            "math": {
                "category_slug": "math",
                "total_points": 1,
                "xp_level": {
                    "level": 2,
                    "stars": 1,
                },
            }
        },
    }


def test_serializable_works_with_inheritance():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    class LocationWithCity(Location):
        city_code = StringType()

        @serializable
        def city_name(self):
            return "Oklahoma" if self.city_code == "OK" else "Unknown"

    location = LocationWithCity(dict(country_code="US", city_code="OK"))

    assert location.country_code == "US"
    assert location.country_name == "United States"
    assert location.city_code == "OK"
    assert location.city_name == "Oklahoma"

    d = location.serialize()
    assert d == {
        "country_code": "US",
        "country_name": "United States",
        "city_code": "OK",
        "city_name": "Oklahoma",
    }


def test_serialize_with_complex_types():
    class QuestionResource(Model):
        url = StringType()

    class Question(Model):
        question_id = StringType(required=True)
        resources = DictType(ListType(ModelType(QuestionResource)))

    q = Question(
        dict(
            question_id="1",
            resources={
                "pictures": [
                    {
                        "url": "http://www.mbl.is",
                    }
                ]
            },
        )
    )

    d = q.serialize()
    assert d == dict(
        question_id="1",
        resources={
            "pictures": [
                {
                    "url": "http://www.mbl.is",
                }
            ]
        },
    )

    q_with_no_resources = Question(dict(question_id="1"))

    d = q_with_no_resources.serialize()
    assert d == dict(
        question_id="1",
        resources=None,
    )


def test_field_with_serialize_when_none():
    class Question(Model):
        id = StringType()
        question = StringType()
        resources = DictType(StringType, serialize_when_none=False)

    q = Question(dict(id=1, question="Who's the man?"))

    d = q.serialize()
    assert d == {
        "id": "1",
        "question": "Who's the man?",
    }

    q = Question(dict(id=1, question="Who's the man?", resources={"A": "B"}))

    d = q.serialize()
    assert d == {"id": "1", "question": "Who's the man?", "resources": {"A": "B"}}


def test_field_with_serialize_when_none_on_outer_only():
    class M(Model):
        listfield = ListType(
            StringType(serialize_when_none=True), serialize_when_none=False
        )
        dictfield = DictType(
            StringType(serialize_when_none=True), serialize_when_none=False
        )

    obj = M()
    obj.listfield = [None]
    obj.dictfield = {"foo": None}
    assert obj.serialize() == {"listfield": [None], "dictfield": {"foo": None}}


def test_field_with_serialize_when_none_on_inner_only():
    class M(Model):
        listfield = ListType(
            StringType(serialize_when_none=False), serialize_when_none=True
        )
        dictfield = DictType(
            StringType(serialize_when_none=False), serialize_when_none=True
        )

    obj = M()
    obj.listfield = [None]
    obj.dictfield = {"foo": None}
    assert obj.serialize() == {"listfield": [], "dictfield": {}}


def test_set_serialize_when_none_on_whole_model():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType)

        class Options:
            serialize_when_none = False

    q = Question(dict(id=1))

    d = q.serialize()
    assert d == {"id": "1"}


def test_possible_to_override_model_wide_serialize_when_none():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType)

        class Options:
            serialize_when_none = False

    class StrictQuestion(Question):
        strictness = IntType()

        class Options:
            serialize_when_none = True

    q = StrictQuestion(dict(id=1))

    d = q.serialize()
    assert d == {"id": "1", "question": None, "resources": None, "strictness": None}


def test_possible_to_override_model_wide_settings_per_field():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType, serialize_when_none=True)

        class Options:
            serialize_when_none = False

    q = Question(dict(id=1))

    d = q.serialize()
    assert d == {"id": "1", "resources": None}


def test_complex_types_hiding_after_apply_role_leaves_it_empty():
    class QuestionResource(Model):
        name = StringType()
        url = StringType()

        class Options:
            serialize_when_none = False
            roles = {"public": whitelist("name")}

    class Question(Model):
        question_id = StringType(required=True)
        resources = DictType(ListType(ModelType(QuestionResource)))

        class Options:
            serialize_when_none = False
            roles = {"public": whitelist("question_id", "resources")}

    q = Question(
        dict(
            question_id="1",
            resources={
                "pictures": [
                    {
                        "url": "http://www.mbl.is",
                    }
                ]
            },
        )
    )

    d = q.serialize("public")
    assert d == {"question_id": "1"}


def test_serialize_none_fields_if_field_says_so():
    class TestModel(Model):
        inst_id = StringType(required=True, serialize_when_none=True)

    q = TestModel({"inst_id": 1})

    d = export_loop(TestModel, q, lambda field, value, context: None)
    assert d == {"inst_id": None}


def test_serialize_none_fields_if_export_loop_says_so():
    class TestModel(Model):
        inst_id = StringType(required=True, serialize_when_none=False)

    q = TestModel({"inst_id": 1})

    d = export_loop(
        TestModel, q, lambda field, value, context: None, export_level=DEFAULT
    )
    assert d == {"inst_id": None}


def test_serialize_print_none_always_gets_you_something():
    class TestModel(Model):
        pass

    q = TestModel()

    d = export_loop(
        TestModel, q, lambda field, value, context: None, export_level=DEFAULT
    )
    assert d == {}


def test_serializable_setter():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

        @country_name.setter
        def country_name(self, value):
            self.country_code = {"United States": "US"}.get(value)

    location = Location()
    location.country_name = "United States"
    assert location.country_code == "US"

    d = location.serialize()
    assert d == {"country_code": "US", "country_name": "United States"}


def test_serializable_setter_override():
    class Player(Model):
        _id = IntType()

        @serializable(IntType())
        def id(self):
            return self._id

        @id.setter
        def id(self, value):
            self._id = value

    p = Player()
    p.id = "1"
    p.validate()

    assert type(1) == type(p.id)
    assert 1 == p.id


def test_serializable_setter_init():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

        @country_name.setter
        def country_name(self, value):
            self.country_code = {"United States": "US"}.get(value)

    location = Location({"country_name": "United States"}, validate=True)
    assert location.country_code == "US"

    d = location.serialize()
    assert d == {"country_code": "US", "country_name": "United States"}


def test_roles_work_with_subclassing():
    class Address(Model):
        private_key = StringType()
        city = StringType()

        class Options:
            roles = {"public": blacklist("private_key")}

    class AddressWithPostalCode(Address):
        postal_code = IntType()

    a = AddressWithPostalCode(
        dict(postal_code=101, city="Reykjavík", private_key="secret")
    )

    d = a.serialize(role="public")
    assert d == {
        "city": "Reykjavík",
        "postal_code": 101,
    }


def test_role_propagate():
    class Address(Model):
        city = StringType()

        class Options:
            roles = {"public": whitelist("city")}

    class User(Model):
        name = StringType(required=True)
        password = StringType()
        addresses = ListType(ModelType(Address))

        class Options:
            roles = {"public": whitelist("name")}

    model = User({"name": "a", "addresses": [{"city": "gotham"}]})
    assert model.addresses[0].city == "gotham"

    d = model.serialize(role="public")
    assert d == {
        "name": "a",
    }


def test_fails_if_role_is_not_found():
    class Player(Model):
        id = StringType()

    p = Player(dict(id="1"))

    with pytest.raises(ValueError):
        p.serialize(role="public")


def test_doesnt_fail_if_role_isnt_found_on_embedded_models():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

        class Options:
            roles = {
                "public": wholelist(),
            }

    class Player(Model):
        id = StringType()
        secret = StringType()

        xp_level = ModelType(ExperienceLevel)

        class Options:
            roles = {"public": blacklist("secret")}

    p = Player(
        dict(id="1", secret="super_secret", xp_level={"level": 1, "title": "Starter"})
    )

    d = p.serialize(role="public")
    assert d == {
        "id": "1",
        "xp_level": {
            "level": 1,
            "title": "Starter",
        },
    }


def test_doesnt_fail_serialize_when_none_on_whole_model_with_roles():
    class Question(Model):
        id = StringType(required=True)
        question = StringType()
        resources = DictType(StringType)

        class Options:
            serialize_when_none = False
            roles = {
                "public": whitelist("id"),
            }

    q = Question({"id": "1"})

    d = q.serialize(role="public")
    assert d == {"id": "1"}


def test_uses_roles_on_embedded_models_if_found():
    class ExperienceLevel(Model):
        level = IntType()
        title = StringType()

        class Options:
            roles = {"public": blacklist("title")}

    class Player(Model):
        id = StringType()
        secret = StringType()

        xp_level = ModelType(ExperienceLevel)

        class Options:
            roles = {"public": blacklist("secret")}

    p = Player(
        dict(id="1", secret="super_secret", xp_level={"level": 1, "title": "Starter"})
    )

    d = p.serialize(role="public")
    assert d == {
        "id": "1",
        "xp_level": {
            "level": 1,
        },
    }


def test_serializable_with_dict_and_roles():
    class Player(Model):
        id = LongType()
        display_name = StringType()

        class Options:
            roles = {"public": blacklist("id")}

    class Game(Model):
        id = StringType()
        result = IntType()
        players = DictType(ModelType(Player), coerce_key=lambda k: int(k))

        class Options:
            roles = {"public": blacklist("result")}

    p1 = Player({"id": 1, "display_name": "A"})
    p2 = Player({"id": 2, "display_name": "B"})

    game = Game({"id": "1", "players": {1: p1, 2: p2}})

    assert game.players[1] == p1
    assert game.players[2] == p2

    d = game.serialize(role="public")

    assert d == {
        "id": "1",
        "players": {
            1: {"display_name": "A"},
            2: {"display_name": "B"},
        },
    }


def test_serializable_with_list_and_roles():
    class Player(Model):
        id = LongType()
        display_name = StringType()

        class Options:
            roles = {"public": blacklist("id")}

    class Game(Model):
        id = StringType()
        result = IntType()
        players = ListType(ModelType(Player))

        class Options:
            roles = {"public": blacklist("result")}

    p1 = Player({"id": 1, "display_name": "A"})
    p2 = Player({"id": 2, "display_name": "B"})

    game = Game({"id": "1", "players": [p1, p2]})

    assert game.players[0] == p1
    assert game.players[1] == p2

    d = game.serialize(role="public")

    assert d == {
        "id": "1",
        "players": [
            {
                "display_name": "A",
            },
            {
                "display_name": "B",
            },
        ],
    }


def test_role_set_operations():

    protected = whitelist("email", "password")
    all_fields = whitelist("id", "name") + protected

    def count(n):
        while True:
            yield n
            n += 1

    class User(Model):
        id = IntType(default=next(count(42)))
        name = StringType()
        email = StringType()
        password = StringType()

        class Options:
            roles = {
                "create": all_fields - ["id"],
                "public": all_fields - ["password"],
                "nospam": blacklist("password") + blacklist("email"),
                "empty": whitelist(),
                "everything": blacklist(),
            }

    roles = User.Options.roles
    assert len(roles["create"]) == 3
    assert len(roles["public"]) == 3
    assert len(roles["nospam"]) == 2
    assert len(roles["empty"]) == 0
    assert len(roles["everything"]) == 0

    # Sets sort different with different Python versions. We should be getting something back
    # like: "whitelist('password', 'email', 'name')"
    s = str(roles["create"])
    assert s.startswith("whitelist(") and s.endswith(")")
    assert sorted(s[10:-1].split(", ")) == ["'email'", "'name'", "'password'"]

    # Similar, but now looking for: <Role whitelist('password', 'email', 'name')>
    r = repr(roles["create"])
    assert r.startswith("<Role whitelist(") and r.endswith(")>")
    assert sorted(r[16:-2].split(", ")) == ["'email'", "'name'", "'password'"]

    data = {
        "id": "NaN",
        "name": "Arthur",
        "email": "adent@hitchhiker.gal",
        "password": "dolphins",
    }

    user = User(
        dict(
            (k, v)
            for k, v in data.items()
            if k in User._schema.options.roles["create"]  # filter by 'create' role
        )
    )

    d = user.serialize(role="public")

    assert d == {
        "id": 42,
        "name": "Arthur",
        "email": "adent@hitchhiker.gal",
    }

    d = user.serialize(role="nospam")

    assert d == {
        "id": 42,
        "name": "Arthur",
    }

    d = user.serialize(role="empty")

    assert d == {}

    d = user.serialize(role="everything")

    assert d == {
        "email": "adent@hitchhiker.gal",
        "id": 42,
        "name": "Arthur",
        "password": "dolphins",
    }


def test_md5_type():
    class M(Model):
        md5 = MD5Type()

    import hashlib

    myhash = hashlib.md5(b"hashthis").hexdigest()
    m = M()
    m.md5 = myhash

    assert m.md5 == myhash
    d = m.serialize()
    assert d == {"md5": myhash}

    m2 = M(d)
    assert m2.md5 == myhash


def test_serializable_with_list_and_default_role():
    class Player(Model):
        id = LongType()
        display_name = StringType()

        class Options:
            roles = {"default": blacklist("id")}

    class Game(Model):
        id = StringType()
        result = IntType()
        players = ListType(ModelType(Player))

        class Options:
            roles = {"default": blacklist("result")}

    p1 = Player({"id": 1, "display_name": "A"})
    p2 = Player({"id": 2, "display_name": "B"})

    game = Game({"id": "1", "players": [p1, p2]})

    assert game.players[0] == p1
    assert game.players[1] == p2

    d = game.serialize(role="default")

    assert d == {
        "id": "1",
        "players": [
            {
                "display_name": "A",
            },
            {
                "display_name": "B",
            },
        ],
    }

    d = game.serialize()

    assert d == {
        "id": "1",
        "players": [
            {
                "display_name": "A",
            },
            {
                "display_name": "B",
            },
        ],
    }


def test_callable_role():
    acl_fields = {
        "user_id_1": ["name"],
        "user_id_2": ["name", "password"],
    }

    class User(Model):
        name = StringType()
        password = StringType()

    u = User({"name": "A", "password": "B"})

    user_1_acl = whitelist(*acl_fields["user_id_1"])
    d = u.serialize(role=user_1_acl)
    assert d == {"name": "A"}

    user_2_acl = whitelist(*acl_fields["user_id_2"])
    d = u.serialize(role=user_2_acl)
    assert d == {"name": "A", "password": "B"}



"""
Django settings for linkify project.

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.6/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "1868fzqv^rh&vh@0$+pa%$b_g241*eq7lmgn*ci#362poo2xby"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "links",
)

MIDDLEWARE_CLASSES = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "links.exception_logger.ConsoleExceptionMiddleware",
)

ROOT_URLCONF = "linkify.urls"

WSGI_APPLICATION = "linkify.wsgi.application"


# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

STATIC_URL = "/static/"



"""This module contains fields that depend on importing `bson`. `bson` is
a part of the pymongo distribution.
"""

import bson  # type: ignore

from ..exceptions import ConversionError
from ..translator import _
from ..types import BaseType

__all__ = ["ObjectIdType"]


class ObjectIdType(BaseType):

    """An field wrapper around MongoDB ObjectIds.  It is correct to say they're
    bson fields, but I am unaware of bson being used outside MongoDB.

    `auto_fill` is disabled by default for ObjectIdType's as they are
    typically obtained after a successful save to Mongo.
    """

    MESSAGES = {
        "convert": _("Couldn't interpret value as an ObjectId."),
    }

    def __init__(self, auto_fill=False, **kwargs):
        self.auto_fill = auto_fill
        super().__init__(**kwargs)

    def to_native(self, value, context=None):
        if not isinstance(value, bson.objectid.ObjectId):
            try:
                value = bson.objectid.ObjectId(str(value))
            except bson.objectid.InvalidId as exc:
                raise ConversionError(self.messages["convert"]) from exc
        return value

    def to_primitive(self, value, context=None):
        return str(value)



import pytest

from schematics.exceptions import DataError
from schematics.models import Model
from schematics.types import BaseType, IntType, StringType
from schematics.types.compound import DictType, ListType, ModelType


def test_nested_mapping():

    mapping = {
        "model_mapping": {
            "modelfield1": {"subfield": "importfield1"},
            "modelfield2": {"subfield": "importfield2"},
        }
    }

    class SubModel(Model):
        subfield = StringType()

    class MainModel(Model):
        modelfield1 = ModelType(SubModel)
        modelfield2 = ModelType(SubModel)

    m1 = MainModel(
        {
            "modelfield1": {"importfield1": "qweasd"},
            "modelfield2": {"importfield2": "qweasd"},
        },
        deserialize_mapping=mapping,
    )

    assert m1.modelfield1.subfield == "qweasd"
    assert m1.modelfield2.subfield == "qweasd"


def test_nested_mapping_with_required():

    mapping = {"model_mapping": {"modelfield": {"subfield": "importfield"}}}

    class SubModel(Model):
        subfield = StringType(required=True)

    class MainModel(Model):
        modelfield = ModelType(SubModel)

    m1 = MainModel(
        {
            "modelfield": {"importfield": "qweasd"},
        },
        partial=False,
        deserialize_mapping=mapping,
    )


def test_submodel_required_field():
    class SubModel(Model):
        req_field = StringType(required=True)
        opt_field = StringType()
        submodelfield = ModelType("SubModel")

    class MainModel(Model):
        intfield = IntType()
        stringfield = StringType()
        modelfield = ModelType(SubModel)

    # By default, model instantiation assumes partial=True
    m1 = MainModel({"modelfield": {"opt_field": "qweasd"}})

    with pytest.raises(DataError):
        m1 = MainModel({"modelfield": {"opt_field": "qweasd"}}, partial=False)

    # Validation implies partial=False
    with pytest.raises(DataError):
        m1.validate()

    m1.validate(partial=True)

    m1 = MainModel({"modelfield": {"req_field": "qweasd"}}, partial=False)

    with pytest.raises(DataError):
        m1 = MainModel(
            {
                "modelfield": {
                    "req_field": "qweasd",
                    "submodelfield": {"opt_field": "qweasd"},
                }
            },
            partial=False,
        )

    m1 = MainModel(
        {
            "modelfield": {
                "req_field": "qweasd",
                "submodelfield": {"req_field": "qweasd"},
            }
        },
        partial=False,
    )


def test_strict_propagation():
    class SubModel(Model):
        subfield = StringType()
        submodelfield = ModelType("SubModel")

    class MainModel(Model):
        modelfield = ModelType(SubModel)

    with pytest.raises(DataError):
        m1 = MainModel(
            {
                "modelfield": {"extrafield": "qweasd"},
            },
            strict=True,
        )

    m1 = MainModel(
        {
            "modelfield": {"extrafield": "qweasd"},
        },
        strict=False,
    )

    assert m1.modelfield._data == {"subfield": None, "submodelfield": None}

    with pytest.raises(DataError):
        m1 = MainModel(
            {
                "modelfield": {"submodelfield": {"extrafield": "qweasd"}},
            },
            strict=True,
        )

    m1 = MainModel(
        {
            "modelfield": {"submodelfield": {"extrafield": "qweasd"}},
        },
        strict=False,
    )



import json

import pytest

from schematics.exceptions import *


def test_error_from_string():

    e = ConversionError("hello")
    assert e.errors == ["hello"]

    e = ValidationError("hello", "world", "!")
    assert e.errors == ["hello", "world", "!"]
    assert len(e) == 3


def _assert(e):
    assert e.errors == ["hello"] and e.errors[0].info == 99


def test_error_from_args():
    _assert(ValidationError("hello", info=99))


def test_error_from_tuple():
    _assert(ValidationError(("hello", 99)))


def test_error_from_message():
    _assert(ValidationError(ErrorMessage("hello", info=99)))


def test_error_from_error():
    _assert(ValidationError(ValidationError(("hello", 99))))


def test_error_from_mixed_args():

    e = ValidationError(
        ("hello", 99),
        "world",
        ErrorMessage("from_msg", info=0),
        ValidationError("from_err", info=1),
    )

    assert e == e.errors == ["hello", "world", "from_msg", "from_err"]
    assert [msg.info for msg in e] == [99, None, 0, 1]


def test_error_from_mixed_list():

    e = ConversionError(
        [
            ("hello", 99),
            "world",
            ErrorMessage("from_msg", info=0),
            ConversionError("from_err", info=1),
        ]
    )

    assert e.errors == ["hello", "world", "from_msg", "from_err"]
    assert [msg.info for msg in e.errors] == [99, None, 0, 1]


def test_error_str():

    assert str(ValidationError("foo")) == '["foo"]'

    e = ValidationError(("foo", None), ("bar", 98), ("baz", [1, 2, 3]))

    assert str(e) == '["foo", "bar: 98", "baz: [1, 2, 3]"]'


def test_error_list_conversion():
    err = ValidationError("A", "B", "C")
    assert list(err) == err.errors


def test_error_eq():
    assert ValidationError("A") == ValidationError("A") == ["A"]
    assert ValidationError("A") != ConversionError("A")
    assert ValidationError("A", "B") == ValidationError("A", "B") == ["A", "B"]
    assert ValidationError("A") != ValidationError("A", "B")


def test_error_message_object():

    assert ErrorMessage("foo") == "foo"
    assert ErrorMessage("foo") != "bar"
    assert ErrorMessage("foo", 1) == ErrorMessage("foo", 1)
    assert ErrorMessage("foo", 1) != ErrorMessage("foo", 2)


@pytest.mark.parametrize(
    "error",
    [
        ErrorMessage("foo", info="bar"),
        BaseError([ErrorMessage("foo", info="bar")]),
        BaseError({"foo": "bar"}),
        ErrorMessage("é"),
        ValidationError("é"),
    ],
)
def test_exception_repr(error):
    assert error == eval(repr(error))


def test_error_failures():

    with pytest.raises(NotImplementedError):
        FieldError()

    with pytest.raises(TypeError):
        ValidationError()

    with pytest.raises(TypeError):
        ValidationError("hello", 99)

    with pytest.raises(TypeError):
        ConversionError(ValidationError("hello"))

    with pytest.raises(TypeError):
        CompoundError(["hello"])


def test_to_primitive():
    error = BaseError(
        {
            "a": [ErrorMessage("a1"), ErrorMessage("a2")],
            "b": {
                "d": ErrorMessage("d_val"),
                "e": ErrorMessage("e_val"),
            },
            "c": ErrorMessage("this is an error"),
        }
    )
    assert error.to_primitive() == {
        "a": ["a1", "a2"],
        "b": {"d": "d_val", "e": "e_val"},
        "c": "this is an error",
    }


def test_to_primitive_list():
    error = BaseError([ErrorMessage("a1"), ErrorMessage("a2")])
    assert error.to_primitive() == ["a1", "a2"]


def test_autopopulate_message_on_none():
    errors = {
        "a": [ErrorMessage("a1"), ErrorMessage("a2")],
        "b": {
            "d": ErrorMessage("d_val"),
            "e": ErrorMessage("e_val"),
        },
        "c": ErrorMessage("this is an error"),
    }
    e = BaseError(errors)
    assert json.loads(str(e)) == BaseError._to_primitive(errors)


@pytest.mark.parametrize(
    "e",
    [
        BaseError(["a", "b"]),
        ConversionError(ErrorMessage("foo"), ErrorMessage("bar")),
        CompoundError({"a": ValidationError(ErrorMessage("foo"))}),
    ],
)
def test_exceptions_is_hashable(e):
    """exceptions must be hashable, as the logging module expects this
    for log.exception()
    """
    hash(e)


@pytest.mark.parametrize("inp,out", [(ConversionError(ErrorMessage("foo")), ["foo"])])
def test_clean_str_representation(inp, out):
    """
    the string representation should be human-readable.  json's format
    provides a legible format for complex data types.
    """
    assert str(inp) == json.dumps(out)



from collections import OrderedDict

import pytest

from schematics.exceptions import (
    ConversionError,
    DataError,
    MockCreationError,
    StopValidationError,
    ValidationError,
)
from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.types.compound import ListType, ModelType


def test_list_field():
    class User(Model):
        ids = ListType(StringType, required=True)

    c = User({"ids": []})

    c.validate({"ids": []})

    assert c.ids == []


def test_list_with_default_type():
    class CategoryStatsInfo(Model):
        slug = StringType()

    class PlayerInfo(Model):
        categories = ListType(ModelType(CategoryStatsInfo))

    math_stats = CategoryStatsInfo(dict(slug="math"))
    twilight_stats = CategoryStatsInfo(dict(slug="twilight"))
    info = PlayerInfo({"categories": [{"slug": "math"}, {"slug": "twilight"}]})

    assert info.categories == [math_stats, twilight_stats]

    d = info.serialize()
    assert d == {
        "categories": [{"slug": "math"}, {"slug": "twilight"}],
    }


def test_set_default():
    class CategoryStatsInfo(Model):
        slug = StringType()

    class PlayerInfo(Model):
        categories = ListType(
            ModelType(CategoryStatsInfo), default=lambda: [], serialize_when_none=True
        )

    info = PlayerInfo()
    assert info.categories == []

    d = info.serialize()
    assert d == {
        "categories": [],
    }


def test_list_defaults_to_none():
    class PlayerInfo(Model):
        following = ListType(StringType)

    info = PlayerInfo()

    assert info.following is None

    assert info.serialize() == {
        "following": None,
    }


def test_list_default_to_none_embedded_model():
    class QuestionResource(Model):
        url = StringType()

    class QuestionResources(Model):
        pictures = ListType(ModelType(QuestionResource))

    class Question(Model):
        id = StringType()
        resources = ModelType(QuestionResources)

    class QuestionPack(Model):
        id = StringType()
        questions = ListType(ModelType(Question))

    question_pack = QuestionPack(
        {
            "id": "1",
            "questions": [
                {
                    "id": "1",
                },
                {
                    "id": "2",
                    "resources": {
                        "pictures": [],
                    },
                },
                {
                    "id": "3",
                    "resources": {
                        "pictures": [
                            {
                                "url": "http://www.mbl.is/djok",
                            }
                        ]
                    },
                },
            ],
        }
    )

    assert question_pack.questions[0].resources is None
    assert question_pack.questions[1].resources["pictures"] == []

    resource = QuestionResource({"url": "http://www.mbl.is/djok"})
    assert question_pack.questions[2].resources["pictures"][0] == resource


def test_validation_with_size_limits():
    class User(Model):
        name = StringType()

    class Card(Model):
        users = ListType(ModelType(User), min_size=1, max_size=2, required=True)

    with pytest.raises(DataError) as exception:
        c = Card({"users": None})
        c.validate()

    assert exception.value.errors["users"] == ["This field is required."]

    with pytest.raises(DataError) as exception:
        c = Card({"users": []})
        c.validate()

    assert exception.value.errors["users"] == ["Please provide at least 1 item."]

    with pytest.raises(DataError) as exception:
        c = Card({"users": [User(), User(), User()]})
        c.validate()

    assert exception.value.errors["users"] == ["Please provide no more than 2 items."]


def test_list_field_required():
    class User(Model):
        ids = ListType(StringType(required=True))

    c = User({"ids": []})

    c.ids = []
    c.validate()

    c.ids = [1]
    c.validate()

    c.ids = [None]
    with pytest.raises(DataError):
        c.validate()


def test_list_field_convert():
    class User(Model):
        ids = ListType(IntType)

    c = User({"ids": ["1", "2"]})

    assert c.ids == [1, 2]


def test_list_coercion():
    field = ListType(StringType)
    assert field(("foobar",)) == ["foobar"]
    assert field(set(("foobar",))) == ["foobar"]
    with pytest.raises(ConversionError):
        field({1: "bar", 2: "baz", 0: "foo"})
    with pytest.raises(ConversionError):
        field("foobar")
    with pytest.raises(ConversionError):
        field(None)


def test_list_model_field():
    class User(Model):
        name = StringType()

    class Card(Model):
        users = ListType(ModelType(User), min_size=1, required=True)

    data = {"users": [{"name": "Doggy"}]}
    c = Card(data)

    c.users = None
    with pytest.raises(DataError) as exception:
        c.validate()

    errors = exception.value.errors
    assert errors["users"] == ["This field is required."]


def test_list_model_field_exception_with_full_message():
    class User(Model):
        name = StringType(max_length=1)

    class Group(Model):
        users = ListType(ModelType(User))

    g = Group({"users": [{"name": "ToLongName"}]})

    with pytest.raises(DataError) as exception:
        g.validate()
    assert exception.value.errors == {
        "users": {0: {"name": ["String value is too long."]}}
    }


def test_compound_fields():
    comments = ListType(ListType, compound_field=StringType)

    assert isinstance(comments.field, ListType)


def test_mock_object():
    assert type(ListType(IntType, required=True).mock()) is list

    assert (
        ListType(
            StringType,
            min_size=0,
            max_size=0,
            required=True,
        ).mock()
        == []
    )

    with pytest.raises(MockCreationError) as exception:
        ListType(IntType, min_size=10, max_size=1, required=True).mock()


def test_mock_object_with_model_type():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)

    assert isinstance(
        ListType(ModelType(User), min_size=1, required=True).mock()[-1], User
    )


def test_issue_453_list_model_field_recursive_import():
    class User(Model):
        name = StringType()

    class Card(Model):
        id = IntType(default=1)
        users = ListType(ModelType(User), min_size=1, required=True)

    card = Card()
    card.users = [
        User({"name": "foo"}),
        User({"name": "bar"}),
        User({"name": "baz"}),
    ]
    card.import_data({"id": "2", "users": [User({"name": "xyz"})]}, recursive=True)
    assert card.serialize() == {"id": 2, "users": [{"name": "xyz"}]}



import json
from collections.abc import Mapping, Sequence
from typing import Optional, Type

from .datastructures import FrozenDict, FrozenList
from .translator import LazyText

__all__ = [
    "BaseError",
    "ErrorMessage",
    "FieldError",
    "ConversionError",
    "ValidationError",
    "StopValidationError",
    "CompoundError",
    "DataError",
    "MockCreationError",
    "UndefinedValueError",
    "UnknownFieldError",
]


class BaseError(Exception):
    def __init__(self, errors):
        """
        The base class for all Schematics errors.

        message should be a human-readable message,
        while errors is a machine-readable list, or dictionary.

        if None is passed as the message, and error is populated,
        the primitive representation will be serialized.

        the Python logging module expects exceptions to be hashable
        and therefore immutable. As a result, it is not possible to
        mutate BaseError's error list or dict after initialization.
        """
        errors = self._freeze(errors)
        super().__init__(errors)

    @property
    def errors(self):
        return self.args[0]

    def to_primitive(self):
        """
        converts the errors dict to a primitive representation of dicts,
        list and strings.
        """
        try:
            return self._primitive
        except AttributeError:
            self._primitive = self._to_primitive(self.errors)
            return self._primitive

    @staticmethod
    def _freeze(obj):
        """freeze common data structures to something immutable."""
        if isinstance(obj, dict):
            return FrozenDict(obj)
        if isinstance(obj, list):
            return FrozenList(obj)
        return obj

    @classmethod
    def _to_primitive(cls, obj):
        """recursive to_primitive for basic data types."""
        if isinstance(obj, str):
            return obj
        if isinstance(obj, Sequence):
            return [cls._to_primitive(e) for e in obj]
        if isinstance(obj, Mapping):
            return dict((k, cls._to_primitive(v)) for k, v in obj.items())
        return str(obj)

    def __str__(self):
        return json.dumps(self.to_primitive())

    def __repr__(self):
        return f"{self.__class__.__name__}({self.errors!r})"

    def __hash__(self):
        return hash(self.errors)

    def __eq__(self, other):
        if type(self) is type(other):
            return self.errors == other.errors
        return self.errors == other

    def __ne__(self, other):
        return not (self == other)


class ErrorMessage:
    def __init__(self, summary, info=None):
        self.type = None
        self.summary = summary
        self.info = info

    def __repr__(self):
        return f"{self.__class__.__name__}({self.summary!r}, {self.info!r})"

    def __str__(self):
        if self.info:
            return f"{self.summary}: {self._info_as_str()}"
        return str(self.summary)

    def _info_as_str(self):
        if isinstance(self.info, int):
            return str(self.info)
        if isinstance(self.info, str):
            return f'"{self.info}"'
        return str(self.info)

    def __eq__(self, other):
        if isinstance(other, ErrorMessage):
            return (
                self.summary == other.summary
                and self.type == other.type
                and self.info == other.info
            )
        if isinstance(other, str):
            return self.summary == other
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.summary, self.type, self.info))


class FieldError(BaseError, Sequence):

    type: Optional[Type[Exception]] = None

    def __init__(self, *args, **kwargs):

        if type(self) is FieldError:
            raise NotImplementedError(
                "Please raise either ConversionError or ValidationError."
            )
        if len(args) == 0:
            raise TypeError("Please provide at least one error or error message.")
        if kwargs:
            items = [ErrorMessage(*args, **kwargs)]
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, list):
                items = list(arg)
            else:
                items = [arg]
        else:
            items = args
        errors = []
        for item in items:
            if isinstance(item, (str, LazyText)):
                errors.append(ErrorMessage(str(item)))
            elif isinstance(item, tuple):
                errors.append(ErrorMessage(*item))
            elif isinstance(item, ErrorMessage):
                errors.append(item)
            elif isinstance(item, self.__class__):
                errors.extend(item.errors)
            else:
                raise TypeError(
                    f"'{type(item).__name__}()' object is neither a {type(self).__name__} nor an error message."
                )
        for error in errors:
            error.type = self.type or type(self)

        super().__init__(errors)

    def __contains__(self, value):
        return value in self.errors

    def __getitem__(self, index):
        return self.errors[index]

    def __iter__(self):
        return iter(self.errors)

    def __len__(self):
        return len(self.errors)


class ConversionError(FieldError, TypeError):
    """Exception raised when data cannot be converted to the correct python type"""


class ValidationError(FieldError, ValueError):
    """Exception raised when invalid data is encountered."""


class StopValidationError(ValidationError):
    """Exception raised when no more validation need occur."""

    type = ValidationError


class CompoundError(BaseError):
    def __init__(self, errors):
        if not isinstance(errors, dict):
            raise TypeError("Compound errors must be reported as a dictionary.")
        for key, value in errors.items():
            if isinstance(value, CompoundError):
                errors[key] = value.errors
            else:
                errors[key] = value
        super().__init__(errors)


class DataError(CompoundError):
    def __init__(self, errors, partial_data=None):
        super().__init__(errors)
        self.partial_data = partial_data


class MockCreationError(ValueError):
    """Exception raised when a mock value cannot be generated."""


class UndefinedValueError(AttributeError, KeyError):
    """Exception raised when accessing a field with an undefined value."""

    def __init__(self, model, name):
        msg = f"'{model.__class__.__name__}' instance has no value for field '{name}'"
        super().__init__(msg)


class UnknownFieldError(KeyError):
    """Exception raised when attempting to access a nonexistent field using the subscription syntax."""

    def __init__(self, model, name):
        msg = f"Model '{model.__class__.__name__}' has no field named '{name}'"
        super().__init__(msg)



import pytest

from schematics.exceptions import DataError, ValidationError
from schematics.models import Model
from schematics.types import IntType, StringType


def test_dont_serialize_untrusted_data():
    """
    Serialization MUST ONLY contain the subset of validated data.
    """

    class Player(Model):
        code = StringType(max_length=4, default=None, serialize_when_none=True)

    p1 = Player({"code": "invalid1"})
    assert p1.serialize() == {"code": None}
    with pytest.raises(DataError):
        p1.validate()
    assert p1.serialize() == {"code": None}


def test_dont_overwrite_with_invalid_data():
    """
    Model-level validators MAY deny data from being set.
    """

    class Player(Model):
        id = IntType()
        name = StringType()

        def validate_id(self, context, value):
            if self._data.valid.get("id"):
                raise ValidationError("Cannot change id")

    p1 = Player({"id": 4})
    p1.validate()
    p1.id = 3
    p1.name = "Douglas"
    with pytest.raises(DataError):
        p1.validate()
    assert p1.id == 4
    assert p1.name == "Douglas"


def test_keep_valid_data_after_multiple_validation():
    """
    Validation MUST maintain a sane state after multiple operations. Valid
    data MUST NEVER be lost or overwritten.
    """

    class Player(Model):
        id = IntType()
        code = StringType(max_length=4)

    p1 = Player({"id": 4})
    p1.validate()
    assert p1.serialize() == {"id": 4, "code": None}
    p1.code = "AAA"
    p1.validate()
    assert p1.serialize() == {"id": 4, "code": "AAA"}
    p1.code = "BBB"
    p1.validate()
    assert p1.serialize() == {"id": 4, "code": "BBB"}
    p1.code = "CCCERR"
    with pytest.raises(DataError):
        p1.validate()
    assert p1.serialize() == {"id": 4, "code": "BBB"}
    p1.validate()
    assert p1.serialize() == {"id": 4, "code": "BBB"}


def test_dont_forget_required_fields_after_multiple_validation():
    """
    Validation MUST NOT forget about required fields when
    clearing untrusted data. Model init data also MUST NOT influence
    the requirement checks.
    """

    class Player(Model):
        code = StringType(required=True)

    p1 = Player()
    with pytest.raises(DataError):
        p1.validate()
    with pytest.raises(DataError):
        p1.validate()



import copy
from functools import partial
from types import FunctionType

from ..exceptions import UndefinedValueError
from ..undefined import Undefined

__all__ = ["calculated", "serializable", "Serializable"]


def serializable(arg=None, **kwargs):
    """A serializable is a way to define dynamic serializable fields that are
    derived from other fields.

    >>> from schematics.models import serializable
    >>> class Location(Model):
    ...     country_code = StringType()
    ...     @serializable
    ...     def country_name(self):
    ...         return {'us': 'United States'}[self.country_code]
    ...
    >>> location = Location({'country_code': 'us'})
    >>> location.serialize()
    {'country_name': 'United States', 'country_code': 'us'}
    >>>
    :param type:
        A custom subclass of `BaseType` for enforcing a certain type
        on serialization.
    :param serialized_name:
        The name of this field in the serialized output.
    """
    from .base import BaseType, TypeMeta

    if isinstance(arg, FunctionType):
        decorator = True
        func = arg
        serialized_type = BaseType
    elif arg is None or isinstance(arg, (BaseType, TypeMeta)):
        decorator = False
        serialized_type = arg or kwargs.pop("type", BaseType)
    else:
        raise TypeError("The argument to 'serializable' must be a function or a type.")

    if isinstance(serialized_type, BaseType):
        # `serialized_type` is already a type instance,
        # so update it with the options found in `kwargs`.
        serialized_type._set_export_level(
            kwargs.pop("export_level", None), kwargs.pop("serialize_when_none", None)
        )
        for name, value in kwargs.items():
            setattr(serialized_type, name, value)
    else:
        serialized_type = serialized_type(**kwargs)

    if decorator:
        return Serializable(type=serialized_type, fget=func)
    return partial(Serializable, type=serialized_type)


def calculated(type, fget, fset=None):
    return Serializable(type=type, fget=fget, fset=fset)


class Serializable:
    def __init__(self, fget, type, fset=None):
        self.type = type
        self.fget = fget
        self.fset = fset

    def __getattr__(self, name):
        return getattr(self.type, name)

    def __get__(self, instance, cls):
        if instance is None:
            return self
        value = self.fget(instance)
        if value is Undefined:
            raise UndefinedValueError(instance, self.name)
        return value

    def __set__(self, instance, value):
        if self.fset is None:
            raise AttributeError(f"can't set attribute {self.name}")
        value = self.type.pre_setattr(value)
        self.fset(instance, value)

    def setter(self, fset):
        self.fset = fset
        return self

    def _repr_info(self):
        return self.type.__class__.__name__

    def __deepcopy__(self, memo):
        return self.__class__(self.fget, type=copy.deepcopy(self.type), fset=self.fset)

    def __repr__(self):
        type_ = f"{self.__class__.__name__}({self._repr_info() or ''}) instance"
        model = f" on {self.owner_model.__name__}" if self.owner_model else ""
        field = f" as '{self.name}'" if self.name else ""
        return f"<{type_}{model}{field}>"



import inspect
from collections import OrderedDict

from ..exceptions import ConversionError
from ..transforms import get_import_context
from ..translator import _
from .base import BaseType

__all__ = ["UnionType"]


def _valid_init_args(type_):
    args = set()
    for cls in type_.__mro__:
        init_args = inspect.getfullargspec(cls.__init__).args[1:]
        args.update(init_args)
        if cls is BaseType:
            break
    return args


def _filter_kwargs(valid_args, kwargs):
    return dict((k, v) for k, v in kwargs.items() if k in valid_args)


class UnionType(BaseType):

    types = None

    MESSAGES = {
        "convert": _("Couldn't interpret value '{0}' as any of {1}."),
    }

    _baseclass_args = _valid_init_args(BaseType)

    def __init__(self, types=None, resolver=None, **kwargs):

        self._types = OrderedDict()
        types = types or self.types
        if resolver:
            self.resolve = resolver

        for type_ in types:
            if isinstance(type_, type) and issubclass(type_, BaseType):
                type_ = type_(**_filter_kwargs(_valid_init_args(type_), kwargs))
            elif not isinstance(type_, BaseType):
                raise TypeError(
                    f"Got '{type_.__class__.__name__}' instance instead of a Schematics type"
                )
            self._types[type_.__class__] = type_
            self.typenames = tuple((cls.__name__ for cls in self._types))

        super().__init__(**_filter_kwargs(self._baseclass_args, kwargs))

    def resolve(self, value, context):
        for field in self._types.values():
            try:
                value = field.convert(value, context)
            except ConversionError:
                pass
            else:
                return field, value
        return None

    def _resolve(self, value, context):
        response = self.resolve(value, context)
        if isinstance(response, type):
            field = self._types[response]
            try:
                response = field, field.convert(value, context)
            except ConversionError:
                pass
        if isinstance(response, tuple):
            return response
        raise ConversionError(self.messages["convert"].format(value, self.typenames))

    def convert(self, value, context=None):
        context = context or get_import_context()
        field, native_value = self._resolve(value, context)
        return native_value

    def validate(self, value, context=None):
        field, _ = self._resolve(value, context)
        return field.validate(value, context)

    def _export(self, value, format, context=None):
        field, _ = self._resolve(value, context)
        return field._export(value, format, context)

    def to_native(self, value, context=None):
        field, _ = self._resolve(value, context)
        return field.to_native(value, context)

    def to_primitive(self, value, context=None):
        field, _ = self._resolve(value, context)
        return field.to_primitive(value, context)



class ConsoleExceptionMiddleware:
    def process_exception(self, request, exception):
        import sys
        import traceback

        exc_info = sys.exc_info()
        print("######################## Exception #############################")
        print("\n".join(traceback.format_exception(*(exc_info or sys.exc_info()))))
        print("################################################################")



import copy
import datetime
import decimal
import itertools
import numbers
import random
import re
import string
import uuid
from collections import OrderedDict
from collections.abc import Iterable
from typing import Any, Optional, Type

from ..common import DEFAULT, NATIVE, NONEMPTY, PRIMITIVE
from ..exceptions import (
    ConversionError,
    MockCreationError,
    StopValidationError,
    ValidationError,
)
from ..translator import _
from ..undefined import Undefined
from ..util import listify
from ..validate import get_validation_context, prepare_validator

__all__ = [
    "BaseType",
    "UUIDType",
    "StringType",
    "MultilingualStringType",
    "NumberType",
    "IntType",
    "LongType",
    "FloatType",
    "DecimalType",
    "HashType",
    "MD5Type",
    "SHA1Type",
    "BooleanType",
    "GeoPointType",
    "DateType",
    "DateTimeType",
    "UTCDateTimeType",
    "TimestampType",
    "TimedeltaType",
]


def fill_template(template, min_length, max_length):
    return template % random_string(
        get_value_in(
            min_length, max_length, padding=len(template) - 2, required_length=1
        )
    )


def get_range_endpoints(min_length, max_length, padding=0, required_length=0):
    if min_length is None:
        min_length = 0
    if max_length is None:
        max_length = max(min_length * 2, 16)

    if padding:
        max_length = max_length - padding
        min_length = max(min_length - padding, 0)

    if max_length < required_length:
        raise MockCreationError("This field is too short to hold the mock data")

    min_length = max(min_length, required_length)
    if max_length < min_length:
        raise MockCreationError("Minimum is greater than maximum")

    return min_length, max_length


def get_value_in(min_length, max_length, padding=0, required_length=0):
    return random.randint(
        *get_range_endpoints(min_length, max_length, padding, required_length)
    )


_alphanumeric = string.ascii_letters + string.digits


def random_string(length, chars=_alphanumeric):
    return "".join(random.choice(chars) for _ in range(length))


_last_position_hint = -1
_next_position_hint = itertools.count()


class TypeMeta(type):

    """
    Meta class for BaseType. Merges `MESSAGES` dict and accumulates
    validator methods.
    """

    def __new__(mcs, name, bases, attrs):
        messages = {}
        validators = OrderedDict()

        for base in reversed(bases):
            try:
                messages.update(base.MESSAGES)
            except AttributeError:
                pass

            try:
                validators.update(base._validators)
            except AttributeError:
                pass

        try:
            messages.update(attrs["MESSAGES"])
        except KeyError:
            pass

        attrs["MESSAGES"] = messages

        for attr_name, attr in attrs.items():
            if attr_name.startswith("validate_"):
                validators[attr_name] = 1
                attrs[attr_name] = prepare_validator(attr, 3)

        attrs["_validators"] = validators

        return type.__new__(mcs, name, bases, attrs)


class BaseType(metaclass=TypeMeta):

    """A base class for Types in a Schematics model. Instances of this
    class may be added to subclasses of ``Model`` to define a model schema.

    Validators that need to access variables on the instance
    can be defined be implementing methods whose names start with ``validate_``
    and accept one parameter (in addition to ``self``)

    :param required:
        Invalidate field when value is None or is not supplied. Default:
        False.
    :param default:
        When no data is provided default to this value. May be a callable.
        Default: None.
    :param serialized_name:
        The name of this field defaults to the class attribute used in the
        model. However if the field has another name in foreign data set this
        argument. Serialized data will use this value for the key name too.
    :param deserialize_from:
        A name or list of named fields for which foreign data sets are
        searched to provide a value for the given field.  This only effects
        inbound data.
    :param choices:
        A list of valid choices. This is the last step of the validator
        chain.
    :param validators:
        A list of callables. Each callable receives the value after it has been
        converted into a rich python type. Default: []
    :param serialize_when_none:
        Dictates if the field should appear in the serialized data even if the
        value is None. Default: None.
    :param messages:
        Override the error messages with a dict. You can also do this by
        subclassing the Type and defining a `MESSAGES` dict attribute on the
        class. A metaclass will merge all the `MESSAGES` and override the
        resulting dict with instance level `messages` and assign to
        `self.messages`.
    :param metadata:
        Dictionary for storing custom metadata associated with the field.
        To encourage compatibility with external tools, we suggest these keys
        for common metadata:
        - *label* : Brief human-readable label
        - *description* : Explanation of the purpose of the field. Used for
          help, tooltips, documentation, etc.
    """

    primitive_type: Optional[Type] = None
    native_type: Optional[Type] = None

    MESSAGES = {
        "required": _("This field is required."),
        "choices": _("Value must be one of {0}."),
    }

    EXPORT_METHODS = {
        NATIVE: "to_native",
        PRIMITIVE: "to_primitive",
    }

    def __init__(
        self,
        required=False,
        default=Undefined,
        serialized_name=None,
        choices=None,
        validators=None,
        deserialize_from=None,
        export_level=None,
        serialize_when_none=None,
        messages=None,
        metadata=None,
    ):
        super().__init__()

        self.required = required
        self._default = default
        self.serialized_name = serialized_name
        if choices and (isinstance(choices, str) or not isinstance(choices, Iterable)):
            raise TypeError('"choices" must be a non-string Iterable')
        self.choices = choices
        self.deserialize_from = listify(deserialize_from)

        self.validators = [
            getattr(self, validator_name) for validator_name in self._validators
        ]
        if validators:
            self.validators += (prepare_validator(func, 2) for func in validators)

        self._set_export_level(export_level, serialize_when_none)

        self.messages = dict(self.MESSAGES, **(messages or {}))
        self.metadata = metadata or {}
        self._position_hint = next(_next_position_hint)  # For ordering of fields

        self.name = None
        self.owner_model = None
        self.parent_field = None
        self.typeclass = self.__class__
        self.is_compound = False

        self.export_mapping = dict(
            (format, getattr(self, fname))
            for format, fname in self.EXPORT_METHODS.items()
        )

    def __repr__(self):
        type_ = f"{self.__class__.__name__}({self._repr_info() or ''}) instance"
        model = f" on {self.owner_model.__name__}" if self.owner_model else ""
        field = f" as '{self.name}'" if self.name else ""
        return f"<{type_}{model}{field}>"

    def _repr_info(self):
        return None

    def __call__(self, value, context=None):
        return self.convert(value, context)

    def __deepcopy__(self, memo):
        return copy.copy(self)

    def _mock(self, context=None):
        return None

    def _setup(self, field_name, owner_model):
        """Perform late-stage setup tasks that are run after the containing model
        has been created.
        """
        self.name = field_name
        self.owner_model = owner_model
        self._input_keys = self._get_input_keys()

    def _set_export_level(self, export_level, serialize_when_none):
        if export_level is not None:
            self.export_level = export_level
        elif serialize_when_none is True:
            self.export_level = DEFAULT
        elif serialize_when_none is False:
            self.export_level = NONEMPTY
        else:
            self.export_level = None

    def get_export_level(self, context):
        if self.owner_model:
            level = self.owner_model._schema.options.export_level
        else:
            level = DEFAULT
        if self.export_level is not None:
            level = self.export_level
        if context.export_level is not None:
            level = context.export_level
        return level

    def get_input_keys(self, mapping=None):
        if mapping:
            return self._get_input_keys(mapping)
        return self._input_keys

    def _get_input_keys(self, mapping=None):
        input_keys = [self.name]
        if self.serialized_name:
            input_keys.append(self.serialized_name)
        if mapping and self.name in mapping:
            input_keys.extend(listify(mapping[self.name]))
        if self.deserialize_from:
            input_keys.extend(self.deserialize_from)
        return input_keys

    @property
    def default(self):
        default = self._default
        if callable(default):
            default = default()
        return default

    def pre_setattr(self, value):
        return value

    def convert(self, value, context=None):
        return self.to_native(value, context)

    def export(self, value, format, context=None):
        return self.export_mapping[format](value, context)

    def to_primitive(self, value, context=None):
        """Convert internal data to a value safe to serialize."""
        return value

    def to_native(self, value, context=None):
        """
        Convert untrusted data to a richer Python construct.
        """
        return value

    def validate(self, value, context=None):
        """
        Validate the field and return a converted value or raise a
        ``ValidationError`` with a list of errors raised by the validation
        chain. Stop the validation process from continuing through the
        validators by raising ``StopValidationError`` instead of ``ValidationError``.

        """
        context = context or get_validation_context()

        if context.convert:
            value = self.convert(value, context)
        elif self.is_compound:
            self.convert(value, context)

        errors = []
        for validator in self.validators:
            try:
                validator(value, context)
            except ValidationError as exc:
                errors.append(exc)
                if isinstance(exc, StopValidationError):
                    break
        if errors:
            raise ValidationError(errors)

        return value

    def check_required(self, value, context):
        if self.required and (value is None or value is Undefined):
            if self.name is None or context and not context.partial:
                raise ConversionError(self.messages["required"])

    def validate_choices(self, value, context):
        if self.choices is not None:
            if value not in self.choices:
                raise ValidationError(
                    self.messages["choices"].format(str(self.choices))
                )

    def mock(self, context=None):
        if not self.required and not random.choice([True, False]):
            return self.default
        if self.choices is not None:
            return random.choice(self.choices)
        return self._mock(context)


class UUIDType(BaseType):

    """A field that stores a valid UUID value."""

    primitive_type = str
    native_type = uuid.UUID

    MESSAGES = {
        "convert": _("Couldn't interpret '{0}' value as UUID."),
    }

    def __init__(self, **kwargs):
        """Create a UUID field."""
        super().__init__(**kwargs)

    def _mock(self, context=None):
        return uuid.uuid4()

    def to_native(self, value, context=None):
        if not isinstance(value, uuid.UUID):
            try:
                value = uuid.UUID(value)
            except (TypeError, ValueError) as exc:
                raise ConversionError(self.messages["convert"].format(value)) from exc
        return value

    def to_primitive(self, value, context=None):
        return str(value)


class StringType(BaseType):

    """A Unicode string field."""

    primitive_type = str
    native_type = str
    allow_casts = (int, bytes)

    MESSAGES = {
        "convert": _("Couldn't interpret '{0}' as string."),
        "decode": _("Invalid UTF-8 data."),
        "max_length": _("String value is too long."),
        "min_length": _("String value is too short."),
        "regex": _("String value did not match validation regex."),
    }

    def __init__(self, regex=None, max_length=None, min_length=None, **kwargs):
        """Create a typing.Text field."""

        self.regex = re.compile(regex) if regex else None
        self.max_length = max_length
        self.min_length = min_length

        super().__init__(**kwargs)

    def _mock(self, context=None):
        return random_string(get_value_in(self.min_length, self.max_length))

    def to_native(self, value, context=None):
        if isinstance(value, str):
            return value
        if isinstance(value, self.allow_casts):
            if isinstance(value, bytes):
                try:
                    return str(value, "utf-8")
                except UnicodeError as exc:
                    raise ConversionError(
                        self.messages["decode"].format(value)
                    ) from exc
            elif isinstance(value, bool):
                pass
            else:
                return str(value)
        raise ConversionError(self.messages["convert"].format(value))

    def validate_length(self, value, context=None):
        length = len(value)
        if self.max_length is not None and length > self.max_length:
            raise ValidationError(self.messages["max_length"])

        if self.min_length is not None and length < self.min_length:
            raise ValidationError(self.messages["min_length"])

    def validate_regex(self, value, context=None):
        if self.regex is not None and self.regex.match(value) is None:
            raise ValidationError(self.messages["regex"])


class NumberType(BaseType):

    """A generic number field.
    Converts to and validates against `number_type` parameter.
    """

    primitive_type: Optional[Type] = None
    native_type: Optional[Type] = None
    number_type: Optional[str] = None
    MESSAGES = {
        "number_coerce": _("Value '{0}' is not {1}."),
        "number_min": _("{0} value should be greater than or equal to {1}."),
        "number_max": _("{0} value should be less than or equal to {1}."),
    }

    def __init__(self, min_value=None, max_value=None, strict=False, **kwargs):
        """Create an int|float field."""

        self.min_value = min_value
        self.max_value = max_value
        self.strict = strict

        super().__init__(**kwargs)

    def _mock(self, context=None):
        number = random.uniform(*get_range_endpoints(self.min_value, self.max_value))
        return self.native_type(number) if self.native_type else number

    def to_native(self, value, context=None):
        if isinstance(value, bool):
            value = int(value)
        if isinstance(value, self.native_type):
            return value
        try:
            native_value = self.native_type(value)
        except (TypeError, ValueError):
            pass
        else:
            if self.native_type is float:  # Float conversion is strict enough.
                return native_value
            if not self.strict and native_value == value:  # Match numeric types.
                return native_value
            if isinstance(value, (str, numbers.Integral)):
                return native_value

        raise ConversionError(
            self.messages["number_coerce"].format(value, self.number_type.lower())
        )

    def validate_range(self, value, context=None):
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(
                self.messages["number_min"].format(self.number_type, self.min_value)
            )

        if self.max_value is not None and value > self.max_value:
            raise ValidationError(
                self.messages["number_max"].format(self.number_type, self.max_value)
            )

        return value


class IntType(NumberType):

    """A field that validates input as an Integer"""

    primitive_type = int
    native_type = int
    number_type = "Int"

    def __init__(self, **kwargs):
        """Create an int field."""
        super().__init__(**kwargs)


LongType = IntType


class FloatType(NumberType):

    """A field that validates input as a Float"""

    primitive_type = float
    native_type = float
    number_type = "Float"

    def __init__(self, **kwargs):
        """Create a float field."""
        super().__init__(**kwargs)


class DecimalType(NumberType):

    """A fixed-point decimal number field."""

    primitive_type = str
    native_type = decimal.Decimal
    number_type = "Decimal"

    def to_primitive(self, value, context=None):
        return str(value)

    def to_native(self, value, context=None):
        if isinstance(value, decimal.Decimal):
            return value

        if not isinstance(value, (str, bool)):
            value = str(value)
        try:
            value = decimal.Decimal(value)
        except (TypeError, decimal.InvalidOperation) as exc:
            raise ConversionError(
                self.messages["number_coerce"].format(value, self.number_type.lower())
            ) from exc

        return value


class HashType(StringType):

    MESSAGES = {
        "hash_length": _("Hash value is wrong length."),
        "hash_hex": _("Hash value is not hexadecimal."),
    }
    LENGTH = -1

    def _mock(self, context=None):
        return random_string(self.LENGTH, string.hexdigits)

    def to_native(self, value, context=None):
        value = super().to_native(value, context)

        if len(value) != self.LENGTH:
            raise ValidationError(self.messages["hash_length"])
        try:
            int(value, 16)
        except ValueError as exc:
            raise ConversionError(self.messages["hash_hex"]) from exc
        return value


class MD5Type(HashType):

    """A field that validates input as resembling an MD5 hash."""

    LENGTH = 32


class SHA1Type(HashType):

    """A field that validates input as resembling an SHA1 hash."""

    LENGTH = 40


class BooleanType(BaseType):

    """A boolean field type. In addition to ``True`` and ``False``, coerces these
    values:

    + For ``True``: "True", "true", "1"
    + For ``False``: "False", "false", "0"

    """

    primitive_type = bool
    native_type = bool

    TRUE_VALUES = ("True", "true", "1")
    FALSE_VALUES = ("False", "false", "0")

    def __init__(self, **kwargs):
        """Create a bool field."""
        super().__init__(**kwargs)

    def _mock(self, context=None):
        return random.choice([True, False])

    def to_native(self, value, context=None):
        if isinstance(value, str):
            if value in self.TRUE_VALUES:
                value = True
            elif value in self.FALSE_VALUES:
                value = False

        elif isinstance(value, int) and value in [0, 1]:
            value = bool(value)

        if not isinstance(value, bool):
            raise ConversionError(_("Must be either true or false."))

        return value


class DateType(BaseType):

    """Defaults to converting to and from ISO8601 date values."""

    primitive_type = str
    native_type = datetime.date

    SERIALIZED_FORMAT = "%Y-%m-%d"
    MESSAGES = {
        "parse": _("Could not parse {0}. Should be ISO 8601 (YYYY-MM-DD)."),
        "parse_formats": _("Could not parse {0}. Valid formats: {1}"),
    }

    def __init__(self, formats=None, **kwargs):
        """Create a datetime.date field."""

        if formats:
            self.formats = listify(formats)
            self.conversion_errmsg = self.MESSAGES["parse_formats"]
        else:
            self.formats = ["%Y-%m-%d"]
            self.conversion_errmsg = self.MESSAGES["parse"]

        self.serialized_format = self.SERIALIZED_FORMAT

        super().__init__(**kwargs)

    def _mock(self, context=None):
        return datetime.date(
            year=random.randrange(600) + 1900,
            month=random.randrange(12) + 1,
            day=random.randrange(28) + 1,
        )

    def to_native(self, value, context=None):
        if isinstance(value, datetime.datetime):
            return value.date()
        if isinstance(value, datetime.date):
            return value

        for fmt in self.formats:
            try:
                return datetime.datetime.strptime(value, fmt).date()
            except (ValueError, TypeError):
                continue

        raise ConversionError(
            self.conversion_errmsg.format(value, ", ".join(self.formats))
        )

    def to_primitive(self, value, context=None):
        return value.strftime(self.serialized_format)


class DateTimeType(BaseType):

    """A field that holds a combined date and time value.

    The built-in parser accepts input values conforming to the ISO 8601 format
    ``<YYYY>-<MM>-<DD>T<hh>:<mm>[:<ss.ssssss>][<z>]``. A space may be substituted
    for the delimiter ``T``. The time zone designator ``<z>`` may be either ``Z``
    or ``±<hh>[:][<mm>]``.

    Values are stored as standard ``datetime.datetime`` instances with the time zone
    offset in the ``tzinfo`` component if available. Raw values that do not specify a time
    zone will be converted to naive ``datetime`` objects unless ``tzd='utc'`` is in effect.

    Unix timestamps are also valid input values and will be converted to UTC datetimes.

    :param formats:
        (Optional) A value or iterable of values suitable as ``datetime.datetime.strptime`` format
        strings, for example ``('%Y-%m-%dT%H:%M:%S', '%Y-%m-%dT%H:%M:%S.%f')``. If the parameter is
        present, ``strptime()`` will be used for parsing instead of the built-in parser.
    :param serialized_format:
        The output format suitable for Python ``strftime``. Default: ``'%Y-%m-%dT%H:%M:%S.%f%z'``
    :param parser:
        (Optional) An external function to use for parsing instead of the built-in parser. It should
        return a ``datetime.datetime`` instance.
    :param tzd:
        Sets the time zone policy.
        Default: ``'allow'``
            ============== ======================================================================
            ``'require'``  Values must specify a time zone.
            ``'allow'``    Values both with and without a time zone designator are allowed.
            ``'utc'``      Like ``allow``, but values with no time zone information are assumed
                           to be in UTC.
            ``'reject'``   Values must not specify a time zone. This also prohibits timestamps.
            ============== ======================================================================
    :param convert_tz:
        Indicates whether values with a time zone designator should be automatically converted to UTC.
        Default: ``False``
            * ``True``:  Convert the datetime to UTC based on its time zone offset.
            * ``False``: Don't convert. Keep the original time and offset intact.
    :param drop_tzinfo:
        Can be set to automatically remove the ``tzinfo`` objects. This option should generally
        be used in conjunction with the ``convert_tz`` option unless you only care about local
        wall clock times. Default: ``False``
            * ``True``:  Discard the ``tzinfo`` components and make naive ``datetime`` objects instead.
            * ``False``: Preserve the ``tzinfo`` components if present.
    """

    primitive_type: Type[Any] = str
    native_type = datetime.datetime

    SERIALIZED_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"

    MESSAGES = {
        "parse": _("Could not parse {0}. Should be ISO 8601 or timestamp."),
        "parse_formats": _("Could not parse {0}. Valid formats: {1}"),
        "parse_external": _("Could not parse {0}."),
        "parse_tzd_require": _("Could not parse {0}. Time zone offset required."),
        "parse_tzd_reject": _("Could not parse {0}. Time zone offset not allowed."),
        "tzd_require": _("Could not convert {0}. Time zone required but not found."),
        "tzd_reject": _("Could not convert {0}. Time zone offsets not allowed."),
        "validate_tzd_require": _("Time zone information required but not found."),
        "validate_tzd_reject": _("Time zone information not allowed."),
        "validate_utc_none": _("Time zone must be UTC but was None."),
        "validate_utc_wrong": _("Time zone must be UTC."),
    }

    REGEX = re.compile(
        r"""
                (?P<year>\d{4})-(?P<month>\d\d)-(?P<day>\d\d)(?:T|\ )
                (?P<hour>\d\d):(?P<minute>\d\d)
                (?::(?P<second>\d\d)(?:(?:\.|,)(?P<sec_frac>\d{1,6}))?)?
                (?:(?P<tzd_offset>(?P<tzd_sign>[+−-])(?P<tzd_hour>\d\d):?(?P<tzd_minute>\d\d)?)
                |(?P<tzd_utc>Z))?$""",
        re.X,
    )

    TIMEDELTA_ZERO = datetime.timedelta(0)

    class fixed_timezone(datetime.tzinfo):
        def utcoffset(self, dt):
            return self.offset

        def fromutc(self, dt):
            return dt + self.offset

        def dst(self, dt):
            return None

        def tzname(self, dt):
            return self.str

        def __str__(self):
            return self.str

        def __repr__(self, info=""):
            return f"{type(self).__name__}({info})"

    class utc_timezone(fixed_timezone):
        offset = datetime.timedelta(0)
        name = str = "UTC"

    class offset_timezone(fixed_timezone):
        def __init__(self, hours=0, minutes=0):
            self.offset = datetime.timedelta(hours=hours, minutes=minutes)
            total_seconds = self.offset.days * 86400 + self.offset.seconds
            self.str = "{0:s}{1:02d}:{2:02d}".format(
                "+" if total_seconds >= 0 else "-",
                int(abs(total_seconds) / 3600),
                int(abs(total_seconds) % 3600 / 60),
            )

        def __repr__(self):
            return DateTimeType.fixed_timezone.__repr__(self, self.str)

    UTC = utc_timezone()
    EPOCH = datetime.datetime(1970, 1, 1, tzinfo=UTC)

    def __init__(
        self,
        formats=None,
        serialized_format=None,
        parser=None,
        tzd="allow",
        convert_tz=False,
        drop_tzinfo=False,
        **kwargs,
    ):
        """Create a datetime.datetime field."""

        if tzd not in ("require", "allow", "utc", "reject"):
            raise ValueError(
                "DateTimeType.__init__() got an invalid value for parameter 'tzd'"
            )
        self.formats = listify(formats)
        self.serialized_format = serialized_format or self.SERIALIZED_FORMAT
        self.parser = parser
        self.tzd = tzd
        self.convert_tz = convert_tz
        self.drop_tzinfo = drop_tzinfo

        super().__init__(**kwargs)

    def _mock(self, context=None):
        dt = datetime.datetime(
            year=random.randrange(600) + 1900,
            month=random.randrange(12) + 1,
            day=random.randrange(28) + 1,
            hour=random.randrange(24),
            minute=random.randrange(60),
            second=random.randrange(60),
            microsecond=random.randrange(1000000),
        )

        if (
            self.tzd == "reject"
            or self.drop_tzinfo
            or self.tzd == "allow"
            and random.randrange(2)
        ):
            return dt
        if self.convert_tz:
            return dt.replace(tzinfo=self.UTC)
        return dt.replace(
            tzinfo=self.offset_timezone(
                hours=random.randrange(-12, 15), minutes=random.choice([0, 30, 45])
            )
        )

    def to_native(self, value, context=None):

        if isinstance(value, datetime.datetime):
            if value.tzinfo is None:
                if not self.drop_tzinfo:
                    if self.tzd == "require":
                        raise ConversionError(
                            self.messages["tzd_require"].format(value)
                        )
                    if self.tzd == "utc":
                        value = value.replace(tzinfo=self.UTC)
            else:
                if self.tzd == "reject":
                    raise ConversionError(self.messages["tzd_reject"].format(value))
                if self.convert_tz:
                    value = value.astimezone(self.UTC)
                if self.drop_tzinfo:
                    value = value.replace(tzinfo=None)
            return value

        if self.formats:
            # Delegate to datetime.datetime.strptime() using provided format strings.
            for fmt in self.formats:
                try:
                    dt = datetime.datetime.strptime(value, fmt)
                    break
                except (ValueError, TypeError):
                    continue
            else:
                raise ConversionError(
                    self.messages["parse_formats"].format(
                        value, ", ".join(self.formats)
                    )
                )
        elif self.parser:
            # Delegate to external parser.
            try:
                dt = self.parser(value)
            except Exception as exc:
                raise ConversionError(
                    self.messages["parse_external"].format(value)
                ) from exc
        else:
            # Use built-in parser.
            try:
                value = float(value)
            except ValueError:
                dt = self.from_string(value)
            except TypeError as exc:
                raise ConversionError(self.messages["parse"].format(value)) from exc
            else:
                dt = self.from_timestamp(value)
            if not dt:
                raise ConversionError(self.messages["parse"].format(value))

        if dt.tzinfo is None:
            if self.tzd == "require":
                raise ConversionError(self.messages["parse_tzd_require"].format(value))
            if self.tzd == "utc" and not self.drop_tzinfo:
                dt = dt.replace(tzinfo=self.UTC)
        else:
            if self.tzd == "reject":
                raise ConversionError(self.messages["parse_tzd_reject"].format(value))
            if self.convert_tz:
                dt = dt.astimezone(self.UTC)
            if self.drop_tzinfo:
                dt = dt.replace(tzinfo=None)

        return dt

    def from_string(self, value):
        match = self.REGEX.match(value)
        if not match:
            return None
        parts = dict(((k, v) for k, v in match.groupdict().items() if v is not None))

        def p(name):
            return int(parts.get(name, 0))

        microsecond = p("sec_frac") and p("sec_frac") * 10 ** (
            6 - len(parts["sec_frac"])
        )
        if "tzd_utc" in parts:
            tz = self.UTC
        elif "tzd_offset" in parts:
            tz_sign = 1 if parts["tzd_sign"] == "+" else -1
            tz_offset = (p("tzd_hour") * 60 + p("tzd_minute")) * tz_sign
            if tz_offset == 0:
                tz = self.UTC
            else:
                tz = self.offset_timezone(minutes=tz_offset)
        else:
            tz = None
        try:
            return datetime.datetime(
                p("year"),
                p("month"),
                p("day"),
                p("hour"),
                p("minute"),
                p("second"),
                microsecond,
                tz,
            )
        except (ValueError, TypeError):
            return None

    def from_timestamp(self, value):
        try:
            return datetime.datetime(1970, 1, 1, tzinfo=self.UTC) + datetime.timedelta(
                seconds=value
            )
        except (ValueError, TypeError):
            return None

    def to_primitive(self, value, context=None):
        if callable(self.serialized_format):
            return self.serialized_format(value)
        return value.strftime(self.serialized_format)

    def validate_tz(self, value, context=None):
        if value.tzinfo is None:
            if not self.drop_tzinfo:
                if self.tzd == "require":
                    raise ValidationError(self.messages["validate_tzd_require"])
                if self.tzd == "utc":
                    raise ValidationError(self.messages["validate_utc_none"])
        else:
            if self.drop_tzinfo:
                raise ValidationError(self.messages["validate_tzd_reject"])
            if self.tzd == "reject":
                raise ValidationError(self.messages["validate_tzd_reject"])
            if self.convert_tz and value.tzinfo.utcoffset(value) != self.TIMEDELTA_ZERO:
                raise ValidationError(self.messages["validate_utc_wrong"])


class UTCDateTimeType(DateTimeType):

    """A variant of ``DateTimeType`` that normalizes everything to UTC and stores values
    as naive ``datetime`` instances. By default sets ``tzd='utc'``, ``convert_tz=True``,
    and ``drop_tzinfo=True``. The standard export format always includes the UTC time
    zone designator ``"Z"``.
    """

    SERIALIZED_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    def __init__(
        self,
        formats=None,
        parser=None,
        tzd="utc",
        convert_tz=True,
        drop_tzinfo=True,
        **kwargs,
    ):
        """Create a datetime.datetime in UTC field."""
        super().__init__(
            formats=formats,
            parser=parser,
            tzd=tzd,
            convert_tz=convert_tz,
            drop_tzinfo=drop_tzinfo,
            **kwargs,
        )


class TimestampType(DateTimeType):

    """A variant of ``DateTimeType`` that exports itself as a Unix timestamp
    instead of an ISO 8601 string. Always sets ``tzd='require'`` and
    ``convert_tz=True``.
    """

    primitive_type = float

    def __init__(self, formats=None, parser=None, drop_tzinfo=False, **kwargs):
        """Create a datetime.datetime as a float field."""
        super().__init__(
            formats=formats,
            parser=parser,
            tzd="require",
            convert_tz=True,
            drop_tzinfo=drop_tzinfo,
            **kwargs,
        )

    def to_primitive(self, value, context=None):
        if value.tzinfo is None:
            value = value.replace(tzinfo=self.UTC)
        else:
            value = value.astimezone(self.UTC)
        delta = value - self.EPOCH
        return delta.total_seconds()


class TimedeltaType(BaseType):

    """Converts Python Timedelta objects into the corresponding value in seconds."""

    primitive_type = float
    native_type = datetime.timedelta

    MESSAGES = {
        "convert": _("Couldn't interpret '{0}' value as Timedelta."),
    }

    DAYS = "days"
    SECONDS = "seconds"
    MICROSECONDS = "microseconds"
    MILLISECONDS = "milliseconds"
    MINUTES = "minutes"
    HOURS = "hours"
    WEEKS = "weeks"

    def __init__(self, precision="seconds", **kwargs):
        """Create a datetime.timedelta field."""
        precision = precision.lower()
        units = (
            self.DAYS,
            self.SECONDS,
            self.MICROSECONDS,
            self.MILLISECONDS,
            self.MINUTES,
            self.HOURS,
            self.WEEKS,
        )
        if precision not in units:
            raise ValueError(
                "TimedeltaType.__init__() got an invalid value for parameter 'precision'"
            )
        self.precision = precision
        super().__init__(**kwargs)

    def _mock(self, context=None):
        return datetime.timedelta(seconds=random.random() * 1000)

    def to_native(self, value, context=None):
        if isinstance(value, datetime.timedelta):
            return value
        try:
            return datetime.timedelta(**{self.precision: float(value)})
        except (ValueError, TypeError) as exc:
            raise ConversionError(self.messages["convert"].format(value)) from exc

    def to_primitive(self, value, context=None):
        base_unit = datetime.timedelta(**{self.precision: 1})
        return int(value.total_seconds() / base_unit.total_seconds())


class GeoPointType(BaseType):

    """A list storing a latitude and longitude."""

    primitive_type = list
    native_type = list

    MESSAGES = {
        "point_min": _("{0} value {1} should be greater than or equal to {2}."),
        "point_max": _("{0} value {1} should be less than or equal to {2}."),
    }

    def _mock(self, context=None):
        return (random.randrange(-90, 90), random.randrange(-180, 180))

    @classmethod
    def _normalize(cls, value):
        if isinstance(value, dict):
            return list(value.values())
        return list(value)

    def to_native(self, value, context=None):
        """Make sure that a geo-value is of type (x, y)"""
        if not isinstance(value, (tuple, list, dict)):
            raise ConversionError(
                _("GeoPointType can only accept tuples, lists, or dicts")
            )
        elements = self._normalize(value)
        if not len(elements) == 2:
            raise ConversionError(_("Value must be a two-dimensional point"))
        if not all(isinstance(v, (float, int)) for v in elements):
            raise ConversionError(_("Both values in point must be float or int"))
        return value

    def validate_range(self, value, context=None):
        latitude, longitude = self._normalize(value)
        if latitude < -90:
            raise ValidationError(
                self.messages["point_min"].format("Latitude", latitude, "-90")
            )
        if latitude > 90:
            raise ValidationError(
                self.messages["point_max"].format("Latitude", latitude, "90")
            )
        if longitude < -180:
            raise ValidationError(
                self.messages["point_min"].format("Longitude", longitude, -180)
            )
        if longitude > 180:
            raise ValidationError(
                self.messages["point_max"].format("Longitude", longitude, 180)
            )


class MultilingualStringType(BaseType):

    """
    A multilanguage string field, stored as a dict with {'locale': 'localized_value'}.

    Minimum and maximum lengths apply to each of the localized values.

    At least one of ``default_locale`` or ``context.app_data['locale']`` must be defined
    when calling ``.to_primitive``.

    """

    primitive_type = str
    native_type = str

    allow_casts = (int, bytes)

    MESSAGES = {
        "convert": _("Couldn't interpret value as string."),
        "max_length": _("String value in locale {0} is too long."),
        "min_length": _("String value in locale {0} is too short."),
        "locale_not_found": _("No requested locale was available."),
        "no_locale": _("No default or explicit locales were given."),
        "regex_locale": _("Name of locale {0} did not match validation regex."),
        "regex_localized": _(
            "String value in locale {0} did not match validation regex."
        ),
    }

    LOCALE_REGEX = r"^[a-z]{2}(:?_[A-Z]{2})?$"

    def __init__(
        self,
        regex=None,
        max_length=None,
        min_length=None,
        default_locale=None,
        locale_regex=LOCALE_REGEX,
        **kwargs,
    ):
        self.regex = re.compile(regex) if regex else None
        self.max_length = max_length
        self.min_length = min_length
        self.default_locale = default_locale
        self.locale_regex = re.compile(locale_regex) if locale_regex else None

        super().__init__(**kwargs)

    def _mock(self, context=None):
        return random_string(get_value_in(self.min_length, self.max_length))

    def to_native(self, value, context=None):
        """Make sure a MultilingualStringType value is a dict or None."""

        if not (value is None or isinstance(value, dict)):
            raise ConversionError(_("Value must be a dict or None"))

        return value

    def to_primitive(self, value, context=None):
        """
        Use a combination of ``default_locale`` and ``context.app_data['locale']`` to return
        the best localized string.

        """
        if value is None:
            return None

        context_locale = None
        if context and "locale" in context.app_data:
            context_locale = context.app_data["locale"]

        # Build a list of all possible locales to try
        possible_locales = []
        for locale in (context_locale, self.default_locale):
            if not locale:
                continue

            if isinstance(locale, str):
                possible_locales.append(locale)
            else:
                possible_locales.extend(locale)

        if not possible_locales:
            raise ConversionError(self.messages["no_locale"])

        for locale in possible_locales:
            if locale in value:
                localized = value[locale]
                break
        else:
            raise ConversionError(self.messages["locale_not_found"])

        if not isinstance(localized, str):
            if isinstance(localized, self.allow_casts):
                if isinstance(localized, bytes):
                    localized = str(localized, "utf-8")
                else:
                    localized = str(localized)
            else:
                raise ConversionError(self.messages["convert"])

        return localized

    def validate_length(self, value, context=None):
        for locale, localized in value.items():
            len_of_value = len(localized) if localized else 0

            if self.max_length is not None and len_of_value > self.max_length:
                raise ValidationError(self.messages["max_length"].format(locale))

            if self.min_length is not None and len_of_value < self.min_length:
                raise ValidationError(self.messages["min_length"].format(locale))

    def validate_regex(self, value, context=None):
        if self.regex is None and self.locale_regex is None:
            return

        for locale, localized in value.items():
            if self.regex is not None and self.regex.match(localized) is None:
                raise ValidationError(self.messages["regex_localized"].format(locale))

            if (
                self.locale_regex is not None
                and self.locale_regex.match(locale) is None
            ):
                raise ValidationError(self.messages["regex_locale"].format(locale))



import copy

import pytest

from schematics.datastructures import Context, DataObject, FrozenDict, FrozenList


def test_data_object_basics():

    d = DataObject({"x": 1, "y": 2})

    assert d == d
    assert d == DataObject(x=1, y=2)
    assert d != DataObject(x=2, y=1)
    assert d != {"x": 1, "y": 2}

    assert DataObject({"f": DataObject({"g": {"x": 1, "y": 2}})}) != {
        "f": {"g": {"x": 1, "y": 2}}
    }

    assert hasattr(d, "x")
    assert "x" in d
    assert not hasattr(d, "z")
    assert "z" not in d

    assert d.x == 1
    with pytest.raises(AttributeError):
        d.z

    assert d["x"] == 1
    with pytest.raises(KeyError):
        d["z"]

    assert d._get("z") is None
    assert d._get("z", 0) == 0

    d.z = 3
    assert d.z == 3
    assert d["z"] == 3

    assert len(d) == 3

    assert (
        set(((k, v) for k, v in d))
        == set(d._items())
        == set((("x", 1), ("y", 2), ("z", 3)))
    )

    assert set((k for k, v in d)) == set(d._keys()) == set(("x", "y", "z"))

    x = d._pop("x")
    assert x == 1 and "x" not in d

    d._clear()
    assert d.__dict__ == {}


def test_data_object_methods():

    d = DataObject({"x": 1})
    d._update({"y": 2})
    d._update(DataObject({"z": 3}, q=4))
    d._update(zip(("n", "m"), (5, 6)))
    assert d == DataObject({"x": 1, "y": 2, "z": 3, "q": 4, "n": 5, "m": 6})

    d = DataObject({"x": 1})
    assert d._setdefault("x") == 1
    a = d._setdefault("a")
    assert a is None and d.a is None
    b = d._setdefault("b", 99)
    assert b == 99 and d.b == 99
    d._setdefaults({"i": 12, "j": 23})
    assert d.i == 12 and d.j == 23

    assert DataObject({"f": DataObject({"g": {"x": 1, "y": 2}})})._to_dict() == {
        "f": {"g": {"x": 1, "y": 2}}
    }


def test_data_object_copy():

    d = DataObject({"f": DataObject({"g": {"x": 1, "y": 2}})})

    d_copy = d._copy()
    assert d_copy == d
    assert id(d_copy) != id(d)
    assert id(d_copy.f) == id(d.f)

    d_deepcopy = copy.deepcopy(d)
    assert d_deepcopy == d
    assert id(d_deepcopy) != id(d)
    assert id(d_deepcopy.f) != id(d.f)


def test_context():
    class FooContext(Context):
        _fields = ("x", "y", "z")

    assert bool(FooContext()) is True

    c = FooContext(x=1, y=2)
    assert c.__dict__ == dict(x=1, y=2)

    with pytest.raises(ValueError):
        FooContext(a=1)

    with pytest.raises(Exception):
        c.x = 0

    c.z = 3
    assert c.__dict__ == dict(x=1, y=2, z=3)

    c = FooContext._new(1, 2, 3)
    assert c.__dict__ == dict(x=1, y=2, z=3)

    with pytest.raises(TypeError):
        FooContext._new(1, 2, 3, 4)

    d = c._branch()
    assert d is c

    d = c._branch(x=None)
    assert d is c

    d = c._branch(x=0)
    assert d is not c
    assert d.__dict__ == dict(x=0, y=2, z=3)

    e = d._branch(x=0)
    assert e is d

    c = FooContext(x=1, y=2)
    c._setdefaults(dict(x=9, z=9))
    assert c.__dict__ == dict(x=1, y=2, z=9)

    c = FooContext(x=1, y=2)
    c._setdefaults(FooContext(x=9, z=9))
    assert c.__dict__ == dict(x=1, y=2, z=9)


def test_frozen_dict():
    frozen_dict = FrozenDict({"foo": "bar"})
    assert frozen_dict["foo"] == "bar"
    assert len(frozen_dict) == 1
    with pytest.raises(TypeError):
        del frozen_dict["foo"]

    with pytest.raises(TypeError):
        frozen_dict["bar"] = "baz"

    assert hash(frozen_dict) == hash(FrozenDict({"foo": "bar"}))
    assert str(frozen_dict) == str({"foo": "bar"}) == repr(frozen_dict)
    assert list(frozen_dict) == list({"foo": "bar"})
    assert frozen_dict == {"foo": "bar"}


def test_frozen_list():
    frozen_list = FrozenList([0, 1, 2])
    assert frozen_list[0] == 0
    assert len(frozen_list) == 3

    with pytest.raises(TypeError):
        del frozen_list[0]

    with pytest.raises(TypeError):
        frozen_list[1] = "baz"

    assert frozen_list == [0, 1, 2]
    assert frozen_list != [0, 1, 3]



from .base import *  # noqa: F401 F403
from .compound import *  # noqa: F401 F403
from .net import *  # noqa: F401 F403
from .serializable import *  # noqa: F401 F403
from .union import *  # noqa: F401 F403



from unittest import mock

from schematics.models import Model
from schematics.types import IntType, ListType, ModelType, StringType


class M(Model):
    intfield = IntType()
    listfield = ListType(ModelType("M"))


class N(M):
    pass


def test_equality_against_mock_any():
    class TestModel(Model):
        pass

    assert TestModel() == mock.ANY


def test_equality_against_same_model():

    m1 = M({"intfield": 1, "listfield": [M({"intfield": 9})]})
    m2 = M({"intfield": 1, "listfield": [M({"intfield": 9})]})
    assert m1 == m2

    # Self-recursive data
    m1.listfield.append(m1)
    m2.listfield.append(m2)
    assert m1 == m2

    m2.listfield[0].intfield = 5
    assert m1 != m2


def test_equality_against_derived_model():

    m = M({"intfield": 1})
    n = N({"intfield": 1})
    assert m != n



import pytest

from schematics.exceptions import DataError
from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.types.compound import ListType, ModelType
from schematics.util import ImportStringError


def test_simple_embedded_models():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p = Player(dict(id=1, location={"country_code": "US"}))

    assert p.id == 1
    assert p.location.country_code == "US"

    p.location = Location({"country_code": "IS"})

    assert isinstance(p.location, Location)
    assert p.location.country_code == "IS"


def test_simple_embedded_models_is_none():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p = Player(dict(id=1))

    assert p.id == 1
    assert p.location is None


def test_simple_embedded_model_set_to_none():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p = Player(dict(id=1))
    p.location = None

    assert p.id == 1
    assert p.location is None


def test_simple_embedded_model_is_none_within_listtype():
    class QuestionResources(Model):
        type = StringType()

    class Question(Model):
        id = StringType()
        resources = ModelType(QuestionResources)

    class QuestionPack(Model):
        id = StringType()
        questions = ListType(ModelType(Question))

    question_pack = QuestionPack(
        {
            "id": "1",
            "questions": [
                {
                    "id": "1",
                },
            ],
        }
    )

    assert question_pack.questions[0].resources is None


def test_raises_validation_error_on_init_with_partial_submodel():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)

    class Card(Model):
        user = ModelType(User)

    u = User({"name": "Arthur"})
    c = Card({"user": u})

    with pytest.raises(DataError):
        c.validate()


def test_model_type():
    class User(Model):
        name = StringType()

    class Card(Model):
        user = ModelType(User)

    c = Card({"user": {"name": "Doggy"}})
    assert isinstance(c.user, User)
    assert c.user.name == "Doggy"


def test_equality_with_embedded_models():
    class Location(Model):
        country_code = StringType()

    class Player(Model):
        id = IntType()
        location = ModelType(Location)

    p1 = Player(dict(id=1, location={"country_code": "US"}))
    p2 = Player(dict(id=1, location={"country_code": "US"}))

    assert id(p1.location) != id(p2.location)
    assert p1.location == p2.location

    assert p1 == p2


def test_default_value_when_embedded_model():
    class Question(Model):
        question_id = StringType(required=True)

        type = StringType(default="text")

    class QuestionPack(Model):

        question = ModelType(Question)

    pack = QuestionPack({"question": {"question_id": 1}})

    assert pack.question.question_id == "1"
    assert pack.question.type == "text"


def test_export_loop_with_subclassed_model():
    class Asset(Model):
        file_name = StringType()

    class S3Asset(Asset):
        bucket_name = StringType()

    class Product(Model):
        title = StringType()
        asset = ModelType(Asset)

    asset = S3Asset({"bucket_name": "assets_bucket", "file_name": "bar"})

    product = Product({"title": "baz", "asset": asset})

    primitive = product.to_primitive()
    assert "bucket_name" in primitive["asset"]

    native = product.to_native()
    assert "bucket_name" in native["asset"]


def test_conversion_error_recursive_overhead():
    conversions = [0]

    class Leaf(Model):
        pass

    next_model = Leaf
    data = "not a mapping"

    for i in range(20):

        class Recursive(Model):
            x = ModelType(next_model, required=True)

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                conversions[0] += 1
                assert conversions[0] < 25

        next_model = Recursive
        data = {"x": data}

    with pytest.raises(DataError):
        next_model(data)


def test_mock_object():
    class User(Model):
        name = StringType(required=True)
        age = IntType(required=True)

    assert ModelType(User, required=True).mock() is not None


def test_specify_model_by_name():
    class M(Model):
        to_one = ModelType("M")
        to_many = ListType(ModelType("M"))
        matrix = ListType(ListType(ModelType("M")))

    assert M.to_one.model_class is M
    assert M.to_many.field.model_class is M
    assert M.matrix.field.field.model_class is M


def test_model_context_pass_to_type():
    from schematics.datastructures import Context
    from schematics.types import BaseType

    class CustomType(BaseType):
        def to_native(self, value, context=None):
            suffix = context.suffix
            return str(value) + suffix

        def to_primitive(self, value, context=None):
            suffix = context.suffix
            return value[: -len(suffix)]

    class Thing(Model):
        x = CustomType()

    context = {"suffix": "z"}
    input = {"x": "thingie"}
    thing = Thing(input, context=context)
    assert thing.x == "thingiez"
    assert thing.to_primitive(context=context) == {"x": "thingie"}
    # try it with a Context object
    model_context = Context(suffix="z!")
    thing2 = Thing(input, context=model_context)
    assert thing2.x == "thingiez!"
    export_context = Context(suffix="z!")
    assert thing2.to_primitive(context=export_context) == {"x": "thingie"}
    with pytest.raises(AttributeError):
        # can't reuse the same Context object as was used for model
        # TODO this may be unexpected to the uninitiated; a custom exception
        #      could explain it better.
        thing2.to_primitive(context=model_context)


def test_model_app_data_pass_to_type():
    from schematics.types import BaseType

    class CustomType(BaseType):
        def to_native(self, value, context=None):
            suffix = context.app_data["suffix"]
            return str(value) + suffix

        def to_primitive(self, value, context=None):
            suffix = context.app_data["suffix"]
            return value[: -len(suffix)]

    class Thing(Model):
        x = CustomType()

    app_data = {"suffix": "z"}
    input = {"x": "thingie"}
    thing = Thing(input, app_data=app_data)
    assert thing.x == "thingiez"
    assert thing.to_primitive(app_data=app_data) == {"x": "thingie"}


class OuterModel:
    class InnerModel(Model):
        test = StringType()


def test_deep_string_search():
    class TestModel(Model):
        deep_model = ModelType("test_model_type.OuterModel.InnerModel")

    test = TestModel(dict(deep_model=dict(test="Abc")))
    assert test.validate() is None

    class TestModel2(Model):
        invalid_model = ModelType("a.c.d.e")

    with pytest.raises(ImportStringError):
        TestModel2(dict(invalid_model=dict(a="1")))


def test_recursive_string_self_reference():
    class RecursiveTestModel(Model):
        recursive_model = ModelType("RecursiveTestModel")
        test = StringType()

    test = RecursiveTestModel(dict(recursive_model=dict(test="Abc")))
    assert test.validate() is None
    assert test.recursive_model.test == "Abc"


def test_circular_string_reference():
    class TestModel1(Model):
        model2 = ModelType("TestModel2")
        name = StringType()

    class TestModel2(Model):
        model1 = ModelType("TestModel1")
        description = StringType()

    data1 = {"name": "Test1"}
    data2 = {"description": "Test2", "model1": data1}

    # TODO: we might want to support locals import late binding someday/somehow
    with pytest.raises(ImportStringError):
        test = TestModel1({"name": "Root", "model2": data2})



from schematics.exceptions import ConversionError, DataError, ValidationError
from schematics.models import Model
from schematics.types import IntType, StringType
from schematics.validate import validate


def test_validate_simple_dict():
    class Player(Model):
        id = IntType()

    validate(Player, {"id": 4})


def test_validate_keep_trusted_data():
    class Player(Model):
        id = IntType()
        name = StringType()

    p1 = Player({"id": 4})
    data = validate(Player, {"name": "Arthur"}, trusted_data=p1._data)

    assert data == {"id": 4, "name": "Arthur"}
    assert data != p1._data


def test_validate_override_trusted_data():
    class Player(Model):
        id = IntType()

    p1 = Player({"id": 4})
    data = validate(Player, {"id": 3}, trusted_data=p1._data)

    assert data == {"id": 3}


def test_validate_ignore_extra_trusted_data():
    class Player(Model):
        id = IntType()

    data = validate(Player, {"id": 4}, trusted_data={"name": "Arthur"})

    assert data == {"id": 4, "name": "Arthur"}


def test_validate_strict_with_trusted_data():
    class Player(Model):
        id = IntType()

    try:
        validate(Player, {"id": 4}, strict=True, trusted_data={"name": "Arthur"})
    except ValidationError as e:
        assert "name" in e.errors


def test_validate_partial_with_trusted_data():
    class Player(Model):
        id = IntType()
        name = StringType(required=True)

    data = validate(Player, {"id": 4}, partial=False, trusted_data={"name": "Arthur"})

    assert data == {"id": 4, "name": "Arthur"}


def test_validate_with_instance_level_validators():
    class Player(Model):
        id = IntType()

        def validate_id(self, data, value, context):
            if self.id:
                raise ValidationError("Cannot change id")

    p1 = Player(trusted_data={"id": 4})

    try:
        validate(Player, p1, {"id": 3})
    except DataError as e:
        assert "id" in e.errors
        assert "Cannot change id" in e.errors["id"]
        assert p1.id == 4



import pytest

from schematics.iteration import Atom, atoms
from schematics.schema import Field, Schema
from schematics.types import IntType, StringType
from schematics.undefined import Undefined


@pytest.fixture
def player_schema():
    schema = Schema(
        "Player",
        Field("id", IntType()),
        Field("first_name", StringType(required=True)),
        Field("last_name", StringType(required=True)),
    )
    return schema


@pytest.fixture
def player_data():
    return {"id": "42", "first_name": "Arthur", "towel": True}


def test_atoms_api_keys_param(player_schema, player_data):
    assert list(atoms(player_schema, player_data, keys=["value"])) == [
        Atom(name=None, field=None, value="42"),
        Atom(name=None, field=None, value="Arthur"),
        Atom(name=None, field=None, value=Undefined),
    ]
    assert list(atoms(player_schema, player_data, keys=["name"])) == [
        Atom(name="id", field=None, value=Undefined),
        Atom(name="first_name", field=None, value=Undefined),
        Atom(name="last_name", field=None, value=Undefined),
    ]



from django.db import models


class Tag(models.Model):
    title = models.CharField(max_length=50)

    def to_dict(self):
        return {"id": self.id, "title": self.title}


class Link(models.Model):
    title = models.CharField(max_length=255)
    url = models.URLField()
    tags = models.ManyToManyField(Tag)

    def to_dict(self):
        return {
            "id": self.id,
            "title": self.title,
            "url": self.url,
            "tags": [tag.to_dict() for tag in self.tags.all()],
        }

    def attach_tags(self, tags):
        for tag in tags:
            self.tags.add(tag)

    def update(self, **kwargs):
        """Update the particular fields rather than all fields.

        This works from Django 1.7+.
        """
        for field, value in kwargs.items():
            setattr(self, field, value)
        self.save(update_fields=list(kwargs.keys()))



import pytest

from schematics.exceptions import DataError
from schematics.models import Model
from schematics.types.base import StringType
from schematics.types.compound import ListType, ModelType
from schematics.types.serializable import serializable


def test_reason_why_we_must_bind_fields():
    class Person(Model):
        name = StringType()

    p1 = Person()
    p2 = Person()

    assert p1 == p2
    assert id(p1) != id(p2)

    assert p1.name == p2.name
    assert id(p1.name) == id(p2.name)

    p1.name = "Jóhann"
    p1.validate()
    p2.validate()

    assert p1 != p2
    assert id(p1) != id(p2)
    assert p1.name != p2.name
    assert id(p1.name) != id(p2.name)

    assert id(p1._schema.fields["name"]) == id(p2._schema.fields["name"])


def test_reason_why_we_must_bind_fields_model_field():
    class Location(Model):
        country_code = StringType(required=True)

    class Person(Model):
        location = ModelType(Location)

    p1 = Person()
    p2 = Person()

    assert p1 == p2
    assert id(p1) != id(p2)

    assert p1.location == p2.location
    assert id(p1.location) == id(p2.location)

    p1.location = {"country_code": "us"}
    assert p1.location.country_code == "us"

    p2.location = {}

    with pytest.raises(DataError):
        p2.validate()

    assert p1 != p2
    assert id(p1) != id(p2)
    assert p1.location != p2.location
    assert id(p1.location) != id(p2.location)

    assert id(p1._schema.fields["location"]) == id(p2._schema.fields["location"])


def test_field_binding():
    class Person(Model):
        name = StringType(required=True)

    class Course(Model):
        id = StringType(required=True, validators=[])
        attending = ListType(ModelType(Person))

    class School(Model):
        courses = ListType(ModelType(Course))

    valid_data = {
        "courses": [
            {"id": "ENG103", "attending": [{"name": "Danny"}, {"name": "Sandy"}]},
            {"id": "ENG203", "attending": [{"name": "Danny"}, {"name": "Sandy"}]},
        ]
    }

    new_school = School(valid_data)
    new_school.validate()

    school = School(valid_data)
    school.validate()

    assert id(new_school) != id(school)

    assert id(new_school.courses[0].attending[0]) != id(school.courses[0].attending[0])


def test_serializable_doesnt_keep_global_state():
    class Location(Model):
        country_code = StringType()

        @serializable
        def country_name(self):
            return "United States" if self.country_code == "US" else "Unknown"

    location_US = Location({"country_code": "US"})
    location_IS = Location({"country_code": "IS"})

    assert (
        location_US._schema.fields["country_name"]
        is location_IS._schema.fields["country_name"]
    )


def test_field_inheritance():
    class A(Model):
        field1 = StringType()
        field2 = StringType()

    class B(A):
        field2 = StringType(required=True)
        field3 = StringType()

    assert A.field1 is A._schema.fields["field1"] is not B._schema.fields["field1"]
    assert A.field2 is A._schema.fields["field2"] is not B._schema.fields["field2"]
    assert "field3" not in A._schema.fields

    assert B.field1 is B._schema.fields["field1"]
    assert B.field2 is B._schema.fields["field2"]
    assert B.field3 is B._schema.fields["field3"]

    assert A.field2.required is False and B.field2.required is True


def test_serializable_inheritance():
    class A(Model):
        @serializable
        def s(self):
            pass

    class B(A):
        pass

    assert A.s is A._schema.fields["s"] is not B._schema.fields["s"]
    assert B.s is B._schema.fields["s"]
    assert A.s.type is not B.s.type
    assert A.s.fget is B.s.fget



import sys
from datetime import datetime, timedelta

import pytest
from dateutil.tz import gettz, tzutc

from schematics.exceptions import ConversionError, ValidationError
from schematics.types import DateTimeType, TimestampType, UTCDateTimeType

UTC = DateTimeType.UTC
NYC = gettz("US/Eastern")
EPOCH = datetime(1970, 1, 1, 0, 0, tzinfo=UTC)


def test_parse_with_defaults():

    field = DateTimeType()

    dt = field.to_native("2015-11-08T12:34")
    assert dt == datetime(2015, 11, 8, 12, 34)

    dt = field.to_native("2015-11-08T12:34:00.1")
    assert dt == datetime(2015, 11, 8, 12, 34, 0, 100000)

    dt = field.to_native("2015-11-08T12:34:56,0369-0730")
    assert dt.utcoffset() == timedelta(hours=-7, minutes=-30)
    assert dt.replace(tzinfo=None) == datetime(2015, 11, 8, 12, 34, 56, 36900)

    assert dt == field.to_native("2015-11-08T12:34:56,0369−0730")  # minus U+2212

    dt = field.to_native("2015-11-08 12:34:56.00200+02:00")
    assert dt.utcoffset() == timedelta(hours=2)
    assert dt.replace(tzinfo=None) == datetime(2015, 11, 8, 12, 34, 56, 2000)

    dt = field.to_native("2015-11-08 12:34:56.00Z")
    assert dt.utcoffset() == timedelta(0)
    assert dt.replace(tzinfo=None) == datetime(2015, 11, 8, 12, 34, 56, 0)


def test_parse_convert():

    field = DateTimeType(convert_tz=True)

    dt = field.to_native("2015-11-08T12:34")
    assert dt == datetime(2015, 11, 8, 12, 34)

    dt = field.to_native("2015-11-08T12:34:56.0369-0730")
    assert dt == datetime(2015, 11, 8, 20, 4, 56, 36900, tzinfo=UTC)

    dt = field.to_native("2015-11-08 12:34:56,00Z")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, 0, tzinfo=UTC)


def test_parse_require_tz():

    field = DateTimeType(tzd="require")

    with pytest.raises(ConversionError):
        dt = field.to_native("2015-11-08 12:34:56.00")

    dt = field.to_native("2015-11-08 12:34:56.00Z")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, tzinfo=UTC)


def test_parse_utc():

    field = DateTimeType(tzd="utc")

    dt = field.to_native("2015-11-08 12:34:56.00")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, tzinfo=UTC)


def test_parse_convert_drop_tzinfo():

    field = DateTimeType(tzd="require", convert_tz=True, drop_tzinfo=True)

    dt = field.to_native("2015-11-08T12:34:56.0369-0730")
    assert dt == datetime(2015, 11, 8, 20, 4, 56, 36900)

    dt = field.to_native("2015-11-08 12:34:56.00Z")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, 0)


def test_parse_reject_tzd():

    field = DateTimeType(tzd="reject")

    with pytest.raises(ConversionError):
        field.to_native("2015-11-08T12:34+0200")

    with pytest.raises(ConversionError):
        field.to_native("2015-11-08T12:34Z")

    dt = field.to_native("2015-11-08T12:34")
    assert dt == datetime(2015, 11, 8, 12, 34)


def test_parse_reject_invalid_type():

    field = DateTimeType()

    with pytest.raises(ConversionError):
        field.to_native(["2017-04-30"])


def test_parse_from_timestamp():

    field = DateTimeType()

    dt = field.to_native("1446991200.7777")
    assert dt == datetime(2015, 11, 8, 14, 00, microsecond=777700, tzinfo=UTC)

    field = DateTimeType(convert_tz=True, drop_tzinfo=True)

    dt = field.to_native("1446991200.7777")
    assert dt == datetime(2015, 11, 8, 14, 00, microsecond=777700)

    dt = field.to_native(1446991200.7777)
    assert dt == datetime(2015, 11, 8, 14, 00, microsecond=777700)


def test_parse_using_formats():

    formats = ("%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%S.%f%z")
    field = DateTimeType(formats=formats)

    dt = field.to_native("2015-11-08T12:34:56.99Z")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, 990000)

    if sys.version_info[0] >= 3:
        # strptime() supports the '%z' directive starting with Python 3.2.
        dt = field.to_native("2015-11-08T12:34:56.099-0700")
        assert dt.utcoffset() == timedelta(hours=-7)
        assert dt.replace(tzinfo=None) == datetime(2015, 11, 8, 12, 34, 56, 99000)


def test_to_native_from_datetime():

    dt_naive_utc = datetime(2015, 6, 1, 14, 00)
    dt_utc = datetime(2015, 6, 1, 14, 00, tzinfo=UTC)
    dt_plustwo = datetime(2015, 6, 1, 16, 00, tzinfo=DateTimeType.offset_timezone(2))
    dt_nyc = datetime(2015, 6, 1, 10, 00, tzinfo=NYC)

    field = DateTimeType(tzd="allow")
    assert field.to_native(dt_naive_utc) == dt_naive_utc
    assert field.to_native(dt_utc) == dt_utc
    assert field.to_native(dt_plustwo) == dt_plustwo
    assert field.to_native(dt_nyc) == dt_nyc

    field = DateTimeType(tzd="utc")
    assert field.to_native(dt_naive_utc) == dt_utc

    field = DateTimeType(tzd="reject")
    assert field.to_native(dt_naive_utc) == dt_naive_utc
    with pytest.raises(ConversionError):
        field.to_native(dt_utc)

    field = DateTimeType(tzd="require")
    assert field.to_native(dt_utc) == dt_utc
    with pytest.raises(ConversionError):
        field.to_native(dt_naive_utc)

    field = DateTimeType(tzd="require", convert_tz=True, drop_tzinfo=True)
    assert field.to_native(dt_naive_utc) == dt_naive_utc

    field = DateTimeType(convert_tz=True)
    assert field.to_native(dt_naive_utc) == dt_naive_utc
    assert field.to_native(dt_utc) == dt_utc
    assert field.to_native(dt_plustwo) == dt_utc
    assert field.to_native(dt_nyc) == dt_utc

    field = DateTimeType(convert_tz=True, drop_tzinfo=True)
    assert field.to_native(dt_naive_utc) == dt_naive_utc
    assert field.to_native(dt_utc) == dt_naive_utc
    assert field.to_native(dt_plustwo) == dt_naive_utc
    assert field.to_native(dt_nyc) == dt_naive_utc


def test_to_primitive():

    dt = datetime(
        2015, 11, 8, 12, 34, 56, 36900, tzinfo=DateTimeType.offset_timezone(-7, -30)
    )

    assert DateTimeType().to_primitive(dt) == "2015-11-08T12:34:56.036900-0730"
    assert (
        DateTimeType(serialized_format="%Y-%m-%d %H:%M:%S").to_primitive(dt)
        == "2015-11-08 12:34:56"
    )


def test_utc_type():

    field = UTCDateTimeType()

    dt = field.to_native("2015-11-08T12:34:56.0369-0730")
    assert dt == datetime(2015, 11, 8, 20, 4, 56, 36900)

    dt = field.to_native("2015-11-08 12:34:56.00Z")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, 0)

    dt = field.to_native("2015-11-08 12:34:56.00")
    assert dt == datetime(2015, 11, 8, 12, 34, 56, 0)

    UTCDateTimeType(default=datetime.utcnow)


def test_timestamp():

    field = TimestampType()
    field_no_tz = TimestampType(drop_tzinfo=True)

    ts = field.to_primitive(datetime(2015, 11, 8, 14, 00))
    assert ts == 1446991200.0
    assert type(ts) == float

    ts = field_no_tz.to_primitive(datetime(2015, 11, 8, 14, 00))
    assert ts == 1446991200.0
    assert type(ts) == float

    ts = field.to_primitive(datetime(2014, 5, 8, 22, 40, 40, tzinfo=gettz("PST8PDT")))
    assert ts == 1399614040.0

    ts = field_no_tz.to_primitive(
        datetime(2014, 5, 8, 22, 40, 40, tzinfo=gettz("PST8PDT"))
    )
    assert ts == 1399614040.0

    dt_with_tz = datetime(2015, 11, 8, 16, 00, tzinfo=DateTimeType.offset_timezone(2))
    dt_no_tz = datetime(2015, 11, 8, 14, 00)
    dt_no_tz_usec = datetime(2015, 11, 8, 14, 00, microsecond=777700)

    assert field.to_primitive(dt_with_tz) == 1446991200.0
    assert field_no_tz.to_primitive(dt_with_tz) == 1446991200.0

    assert field.to_native(dt_with_tz) == dt_no_tz.replace(tzinfo=UTC)
    assert field_no_tz.to_native(dt_with_tz) == dt_no_tz

    assert field.to_native(1446991200.7777) == dt_no_tz_usec.replace(tzinfo=UTC)
    assert field_no_tz.to_native(1446991200.7777) == dt_no_tz_usec

    with pytest.raises(ConversionError):
        field.to_native(dt_no_tz)

    ts = field.to_primitive(dt_no_tz_usec)
    assert ts == 1446991200.7777
    assert type(ts) == float

    ts = field.to_primitive(datetime(2014, 5, 8, 22, 40, 40, tzinfo=UTC))
    assert ts == 1399588840.0

    assert field.to_primitive(EPOCH) == 0
    assert field.to_native(0) == EPOCH


def test_validate_tz():

    dt_naive = lambda: datetime(2015, 6, 1, 14, 00)
    dt_utc = lambda: datetime(2015, 6, 1, 14, 00, tzinfo=UTC)
    dt_plustwo = lambda: datetime(
        2015, 6, 1, 16, 00, tzinfo=DateTimeType.offset_timezone(2)
    )
    dt_nyc = lambda: datetime(2015, 6, 1, 10, 00, tzinfo=NYC)

    all_values = [dt_naive, dt_utc, dt_plustwo, dt_nyc]

    def test_field(field, valid):
        for value in all_values:
            if value in valid:
                field.validate_tz(value())
            else:
                with pytest.raises(ValidationError):
                    field.validate_tz(value())

    test_field(DateTimeType(tzd="allow"), [dt_naive, dt_utc, dt_plustwo, dt_nyc])
    test_field(DateTimeType(convert_tz=True), [dt_naive, dt_utc])
    test_field(DateTimeType(tzd="utc"), [dt_utc, dt_plustwo, dt_nyc])
    test_field(DateTimeType(tzd="utc", convert_tz=True), [dt_utc])
    test_field(DateTimeType(tzd="reject"), [dt_naive])
    test_field(DateTimeType(tzd="require"), [dt_utc, dt_plustwo, dt_nyc])
    test_field(DateTimeType(drop_tzinfo=True), [dt_naive])


@pytest.mark.parametrize("tzd", ("require", "allow", "utc", "reject"))
@pytest.mark.parametrize("convert_tz", (True, False))
@pytest.mark.parametrize("drop_tzinfo", (True, False))
def test_mock(tzd, convert_tz, drop_tzinfo):

    field = DateTimeType(tzd=tzd, convert_tz=convert_tz, drop_tzinfo=drop_tzinfo)
    for _ in range(20):
        field.validate(field._mock())



import pytest
from bson.objectid import ObjectId

from schematics.contrib.mongo import ObjectIdType
from schematics.exceptions import ConversionError

FAKE_OID = ObjectId()


def test_to_native():
    oid = ObjectIdType()

    assert oid.to_native(FAKE_OID) == FAKE_OID
    assert oid.to_native(str(FAKE_OID)) == FAKE_OID

    with pytest.raises(ConversionError):
        oid.to_native("foo")


def test_to_primitive():
    oid = ObjectIdType()

    assert oid.to_primitive(FAKE_OID) == str(FAKE_OID)
    assert oid.to_primitive(str(FAKE_OID)) == str(FAKE_OID)


def test_validate_id():
    oid = ObjectIdType()

    oid.validate(FAKE_OID)
    oid.validate(str(FAKE_OID))

    with pytest.raises(ConversionError):
        oid.validate("foo")



"""
A type and singleton value (like None) to represent fields that
have not been initialized.
"""


class UndefinedType:

    _instance = None

    def __str__(self):
        return "Undefined"

    def __repr__(self):
        return "Undefined"

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def __lt__(self, other):
        self._cmp_err(other, "<")

    def __gt__(self, other):
        self._cmp_err(other, ">")

    def __le__(self, other):
        self._cmp_err(other, "<=")

    def __ge__(self, other):
        self._cmp_err(other, ">=")

    def _cmp_err(self, other, op):
        raise TypeError(
            f"unorderable types: {self.__class__.__name__}() {op} {other.__class__.__name__}()"
        )

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls)
        elif cls is not UndefinedType:
            raise TypeError("type 'UndefinedType' is not an acceptable base type")
        return cls._instance

    def __init__(self):
        pass

    def __setattr__(self, name, value):
        raise TypeError("'UndefinedType' object does not support attribute assignment")


Undefined = UndefinedType()



import itertools
import types
from collections import OrderedDict
from typing import List

from .common import DEFAULT, DROP, NATIVE, NONEMPTY, NOT_NONE, PRIMITIVE
from .datastructures import Context
from .exceptions import CompoundError, DataError, FieldError
from .iteration import atom_filter, atoms
from .role import Role
from .undefined import Undefined
from .util import listify

__all__: List[str] = []


###
# Transform loops
###


def schema_from(obj):
    try:
        return obj._schema
    except AttributeError:
        return obj


def import_loop(
    schema,
    mutable,
    raw_data=None,
    field_converter=None,
    trusted_data=None,
    mapping=None,
    partial=False,
    strict=False,
    init_values=False,
    apply_defaults=False,
    convert=True,
    validate=False,
    new=False,
    oo=False,
    recursive=False,
    app_data=None,
    context=None,
):
    """
    The import loop is designed to take untrusted data and convert it into the
    native types, as described in ``schema``.  It does this by calling
    ``field_converter`` on every field.

    Errors are aggregated and returned by throwing a ``ModelConversionError``.

    :param schema:
        The Schema to use as source for validation.
    :param mutable:
        A mapping or instance that can be changed during validation by Schema
        functions.
    :param raw_data:
        A mapping to be converted into types according to ``schema``.
    :param field_converter:
        This function is applied to every field found in ``instance_or_dict``.
    :param trusted_data:
        A ``dict``-like structure that may contain already validated data.
    :param partial:
        Allow partial data to validate; useful for PATCH requests.
        Essentially drops the ``required=True`` arguments from field
        definitions. Default: False
    :param strict:
        Complain about unrecognized keys. Default: False
    :param apply_defaults:
        Whether to set fields to their default values when not present in input data.
    :param app_data:
        An arbitrary container for application-specific data that needs to
        be available during the conversion.
    :param context:
        A ``Context`` object that encapsulates configuration options and ``app_data``.
        The context object is created upon the initial invocation of ``import_loop``
        and is then propagated through the entire process.
    """
    if raw_data is None:
        raw_data = mutable
    got_data = raw_data is not None

    context = Context._make(context)
    try:
        context.initialized
    except Exception:
        if isinstance(field_converter, types.FunctionType):
            field_converter = BasicConverter(field_converter)
        context._setdefaults(
            {
                "initialized": True,
                "field_converter": field_converter,
                "trusted_data": trusted_data or {},
                "mapping": mapping or {},
                "partial": partial,
                "strict": strict,
                "init_values": init_values,
                "apply_defaults": apply_defaults,
                "convert": convert,
                "validate": validate,
                "new": new,
                "oo": oo,
                "recursive": recursive,
                "app_data": app_data if app_data is not None else {},
            }
        )

    raw_data = context.field_converter.pre(schema, raw_data, context)

    _field_converter = context.field_converter
    _model_mapping = context.mapping.get("model_mapping")

    data = dict(context.trusted_data) if context.trusted_data else {}
    errors = {}

    if got_data and context.validate:
        errors = _mutate(schema, mutable, raw_data, context)

    if got_data:
        # Determine all acceptable field input names
        all_fields = schema_from(schema).valid_input_keys
        if context.mapping:
            mapped_keys = set(
                itertools.chain(
                    *(
                        listify(input_keys)
                        for target_key, input_keys in context.mapping.items()
                        if target_key != "model_mapping"
                    )
                )
            )
            all_fields = all_fields | mapped_keys
        if context.strict:
            # Check for rogues if strict is set
            rogue_fields = set(raw_data) - all_fields
            if rogue_fields:
                for field in rogue_fields:
                    errors[field] = "Rogue field"

    atoms_filter = None
    if not context.validate:
        # optimization: convert without validate doesn't require to touch setters
        atoms_filter = atom_filter.not_setter
    for field_name, field, value in atoms(schema, raw_data, filter=atoms_filter):
        serialized_field_name = field.serialized_name or field_name

        if got_data and value is Undefined:
            for key in field.get_input_keys(context.mapping):
                if key and key != field_name and key in raw_data:
                    value = raw_data[key]
                    break

        if value is Undefined:
            if field_name in data:
                continue
            if context.apply_defaults:
                value = field.default
        if value is Undefined and context.init_values:
            value = None

        if got_data:
            if field.is_compound:
                if context.trusted_data and context.recursive:
                    td = context.trusted_data.get(field_name)
                    if not all(hasattr(td, attr) for attr in ("keys", "__getitem__")):
                        td = {field_name: td}
                else:
                    td = {}
                if _model_mapping:
                    submap = _model_mapping.get(field_name)
                else:
                    submap = {}
                field_context = context._branch(trusted_data=td, mapping=submap)
            else:
                field_context = context
            try:
                value = _field_converter(field, value, field_context)
            except (FieldError, CompoundError) as exc:
                errors[serialized_field_name] = exc
                if context.apply_defaults:
                    value = field.default
                    if value is not Undefined:
                        data[field_name] = value
                if isinstance(exc, DataError):
                    data[field_name] = exc.partial_data
                continue

        if value is Undefined:
            continue

        data[field_name] = value

    if not context.validate:
        for field_name, field, value in atoms(
            schema, raw_data, filter=atom_filter.has_setter
        ):
            data[field_name] = value

    if errors:
        raise DataError(errors, data)

    data = context.field_converter.post(schema, data, context)

    return data


def _mutate(schema, mutable, raw_data, context):
    """
    Mutates the converted data before validation. Allows Schema fields to modify
    and create data values on mutable.
    """
    errors = {}
    for field_name, field, value in atoms(
        schema, raw_data, filter=atom_filter.has_setter
    ):
        if value is Undefined:
            continue
        try:
            value = context.field_converter(field, value, context)
            field.__set__(mutable, value)
        except (FieldError, CompoundError) as exc:
            serialized_field_name = field.serialized_name or field_name
            errors[serialized_field_name] = exc
            continue
        except AttributeError:
            pass
    raw_data.update((key, mutable[key]) for key in mutable)
    return errors


def export_loop(
    schema,
    instance_or_dict,
    field_converter=None,
    role=None,
    raise_error_on_role=True,
    export_level=None,
    app_data=None,
    context=None,
):
    """
    The export_loop function is intended to be a general loop definition that
    can be used for any form of data shaping, such as application of roles or
    how a field is transformed.

    :param schema:
        The Schema to use as source for validation.
    :param instance_or_dict:
        The structure where fields from schema are mapped to values. The only
        expectation for this structure is that it implements a ``dict``
        interface.
    :param field_converter:
        This function is applied to every field found in ``instance_or_dict``.
    :param role:
        The role used to determine if fields should be left out of the
        transformation.
    :param raise_error_on_role:
        This parameter enforces strict behavior which requires substructures
        to have the same role definition as their parent structures.
    :param app_data:
        An arbitrary container for application-specific data that needs to
        be available during the conversion.
    :param context:
        A ``Context`` object that encapsulates configuration options and ``app_data``.
        The context object is created upon the initial invocation of ``import_loop``
        and is then propagated through the entire process.
    """
    context = Context._make(context)
    try:
        context.initialized
    except Exception:
        if isinstance(field_converter, types.FunctionType):
            field_converter = BasicConverter(field_converter)
        context._setdefaults(
            {
                "initialized": True,
                "field_converter": field_converter,
                "role": role,
                "raise_error_on_role": raise_error_on_role,
                "export_level": export_level,
                "app_data": app_data if app_data is not None else {},
            }
        )

    instance_or_dict = context.field_converter.pre(schema, instance_or_dict, context)

    if schema_from(schema).options.export_order:
        data = OrderedDict()
    else:
        data = {}

    filter_func = (
        context.role
        if callable(context.role)
        else schema_from(schema).options.roles.get(context.role)
    )
    if filter_func is None:
        if context.role and context.raise_error_on_role:
            raise ValueError(f'{schema.name} Model has no role "{context.role}"')
        filter_func = schema_from(schema).options.roles.get("default")

    _field_converter = context.field_converter

    for field_name, field, value in atoms(schema, instance_or_dict):
        serialized_name = field.serialized_name or field_name

        if filter_func is not None and filter_func(field_name, value):
            continue

        _export_level = field.get_export_level(context)

        if _export_level == DROP:
            continue

        if value is not None and value is not Undefined:
            value = _field_converter(field, value, context)

        if value is Undefined:
            if _export_level <= DEFAULT:
                continue
        elif value is None:
            if _export_level <= NOT_NONE:
                continue
        elif field.is_compound and len(value) == 0:
            if _export_level <= NONEMPTY:
                continue

        if value is Undefined:
            value = None

        data[serialized_name] = value

    data = context.field_converter.post(schema, data, context)

    return data


###
# Field filtering
###


def wholelist(*field_list):
    """
    Returns a function that evicts nothing. Exists mainly to be an explicit
    allowance of all fields instead of a using an empty blacklist.
    """
    return Role(Role.wholelist, field_list)


def whitelist(*field_list):
    """
    Returns a function that operates as a whitelist for the provided list of
    fields.

    A whitelist is a list of fields explicitly named that are allowed.
    """
    return Role(Role.whitelist, field_list)


def blacklist(*field_list):
    """
    Returns a function that operates as a blacklist for the provided list of
    fields.

    A blacklist is a list of fields explicitly named that are not allowed.
    """
    return Role(Role.blacklist, field_list)


###
# Field converter interface
###


class Converter:
    def __call__(self, field, value, context):
        raise NotImplementedError

    def pre(self, model_class, instance_or_dict, context):
        return instance_or_dict

    def post(self, model_class, data, context):
        return data


class BasicConverter(Converter):
    def __init__(self, func):
        self.func = func

    def __call__(self, *args):
        return self.func(*args)


###
# Standard export converters
###


@BasicConverter
def to_native_converter(field, value, context):
    return field.export(value, NATIVE, context)


@BasicConverter
def to_primitive_converter(field, value, context):
    return field.export(value, PRIMITIVE, context)


###
# Standard import converters
###


@BasicConverter
def import_converter(field, value, context):
    field.check_required(value, context)
    if value is None or value is Undefined:
        return value
    return field.convert(value, context)


@BasicConverter
def validation_converter(field, value, context):
    field.check_required(value, context)
    if value is None or value is Undefined:
        return value
    return field.validate(value, context)


###
# Context stub factories
###


def get_import_context(field_converter=import_converter, **options):
    import_options = {
        "field_converter": field_converter,
        "partial": False,
        "strict": False,
        "convert": True,
        "validate": False,
        "new": False,
        "oo": False,
    }
    import_options.update(options)
    return Context(**import_options)


def get_export_context(field_converter=to_native_converter, **options):
    export_options = {"field_converter": field_converter, "export_level": None}
    export_options.update(options)
    return Context(**export_options)


###
# Import and export functions
###


def convert(cls, mutable, raw_data=None, **kwargs):
    return import_loop(cls, mutable, raw_data, import_converter, **kwargs)


def to_native(cls, instance_or_dict, **kwargs):
    return export_loop(cls, instance_or_dict, to_native_converter, **kwargs)


def to_primitive(cls, instance_or_dict, **kwargs):
    return export_loop(cls, instance_or_dict, to_primitive_converter, **kwargs)



import json

from django.test import Client, TestCase


class TestCreateLink(TestCase):
    def setUp(self):
        super().setUp()
        self.client = Client()

    def test_create_with_bad_data(self):
        data = {
            "name": "Django",
            "url": "https://github.com/django/django",
            "tags": ["Web Framework", "Python"],
        }
        resp = self.client.post(
            "/links/", content_type="application/json", data=json.dumps(data)
        )

        assert resp.status_code == 400

    def test_create(self):
        data = {
            "title": "Django",
            "url": "https://github.com/django/django",
            "tags": ["Web Framework", "Python"],
        }
        resp = self.client.post(
            "/links/", content_type="application/json", data=json.dumps(data)
        )
        obj = json.loads(resp.content.decode())

        assert resp.status_code == 201
        assert obj["id"]

        assert len(obj["tags"]) == 2
        assert obj["title"] == data["title"]
        assert obj["url"] == data["url"]


class TestUpdateLink(TestCase):
    def setUp(self):
        super().setUp()
        self.client = Client()

        data = {
            "title": "Django",
            "url": "https://github.com/django/django",
            "tags": ["Web Framework", "Python"],
        }
        resp = self.client.post(
            "/links/", content_type="application/json", data=json.dumps(data)
        )
        self.link = json.loads(resp.content.decode())

    def test_update(self):
        data = {"title": "Flask", "url": "https://github.com/mitsuhiko/flask"}
        pk = self.link["id"]

        resp = self.client.patch(
            "/links/{}/".format(pk),
            content_type="application/json",
            data=json.dumps(data),
        )

        obj = json.loads(resp.content.decode())

        assert resp.status_code == 202
        assert obj["title"] == data["title"]
        assert obj["url"] == data["url"]

    def test_update_with_bad_params(self):
        data = {"name": "Flask"}

        pk = self.link["id"]

        resp = self.client.patch(
            "/links/{}/".format(pk),
            content_type="application/json",
            data=json.dumps(data),
        )

        assert resp.status_code == 400


class TestListLink(TestCase):
    def setUp(self):
        super().setUp()
        self.client = Client()

        data = {
            "title": "Django",
            "url": "https://github.com/django/django",
            "tags": ["Web Framework", "Python"],
        }
        resp = self.client.post(
            "/links/", content_type="application/json", data=json.dumps(data)
        )
        self.link = json.loads(resp.content.decode())

    def test_list(self):
        resp = self.client.get("/links/")
        obj = json.loads(resp.content.decode())

        assert resp.status_code == 200
        assert int(obj["total"]) == 1
        for item in obj["items"]:
            assert item["id"]
            assert item["title"]
            assert item["url"]

            for tag in item["tags"]:
                assert tag["id"]
                assert tag["title"]


class TestReadLink(TestCase):
    def setUp(self):
        super().setUp()
        self.client = Client()

        self.data = {
            "title": "Django",
            "url": "https://github.com/django/django",
            "tags": ["Web Framework", "Python"],
        }
        resp = self.client.post(
            "/links/", content_type="application/json", data=json.dumps(self.data)
        )
        self.link = json.loads(resp.content.decode())

    def test_read(self):
        resp = self.client.get("/links/{}/".format(self.link["id"]))
        obj = json.loads(resp.content.decode())

        assert obj["id"] == 1
        assert obj["title"] == self.data["title"]
        assert obj["url"] == self.data["url"]

        for tag in obj["tags"]:
            assert tag["id"]
            assert tag["title"] in self.data["tags"]

    def test_missing(self):
        resp = self.client.get("/links/23/")

        assert resp.status_code == 404


class TestDeleteLink(TestCase):
    def setUp(self):
        super().setUp()
        self.client = Client()

        self.data = {
            "title": "Django",
            "url": "https://github.com/django/django",
            "tags": ["Web Framework", "Python"],
        }
        resp = self.client.post(
            "/links/", content_type="application/json", data=json.dumps(self.data)
        )
        self.link = json.loads(resp.content.decode())

    def test_delete(self):
        resp = self.client.delete("/links/{}/".format(self.link["id"]))

        assert resp.status_code == 204

        resp = self.client.get("/links/{}/".format(self.link["id"]))
        assert resp.status_code == 404



class LazyText:
    def __init__(self, message):
        self.message = message

    def __str__(self):
        translator = _.real_translator
        return translator(self.message) if translator else self.message

    def __mod__(self, other):
        return str(self) % other

    def format(self, *args, **kwargs):
        return str(self).format(*args, **kwargs)


class Translator:
    """A placeholder which could call a function like lazy_gettext and make messages translatable."""

    def __init__(self):
        self.real_translator = None

    def __call__(self, message, lazy=True, *args, **kwargs):
        return LazyText(message) if lazy else str(LazyText(message))

    def register_translator(self, new_translator):
        self.real_translator = new_translator


_ = Translator()
register_translator = _.register_translator



"""
Define constants and expose the compatibility overrides to all modules.
"""

from .util import Constant

NATIVE = Constant("NATIVE", 0)
PRIMITIVE = Constant("PRIMITIVE", 1)

DROP = Constant("DROP", 0)
NONEMPTY = Constant("NONEMPTY", 1)
NOT_NONE = Constant("NOT_NONE", 2)
DEFAULT = Constant("DEFAULT", 10)
ALL = Constant("ALL", 99)


__all__ = ["NATIVE", "PRIMITIVE", "DROP", "NONEMPTY", "NOT_NONE", "DEFAULT", "ALL"]



import inspect
import itertools
from collections import OrderedDict

from .common import DEFAULT, NONEMPTY


class Schema:
    def __init__(self, name, *fields, **kw):
        self.name = name
        self.model = kw.get("model", None)
        self.options = kw.get("options", SchemaOptions())
        self.validators = kw.get("validators", {})
        self.fields = OrderedDict()
        for field in fields:
            self.append_field(field)

    @property
    def valid_input_keys(self):
        return set(itertools.chain(*(t.get_input_keys() for t in self.fields.values())))

    def append_field(self, field):
        self.fields[field.name] = field.type
        field.type._setup(field.name, self.model)  # TODO: remove model reference


class SchemaOptions:
    def __init__(
        self,
        namespace=None,
        roles=None,
        export_level=DEFAULT,
        serialize_when_none=None,
        export_order=False,
        extras=None,
    ):
        self.namespace = namespace
        self.roles = roles or {}
        self.export_level = export_level
        if serialize_when_none is True:
            self.export_level = DEFAULT
        elif serialize_when_none is False:
            self.export_level = NONEMPTY
        self.export_order = export_order
        self.extras = extras or {}

        for key, value in self.extras.items():
            setattr(self, key, value)

    def __iter__(self):
        for key, value in inspect.getmembers(self):
            if not key.startswith("_"):
                yield key, value


class Field:

    __slots__ = ("name", "type")

    def __init__(self, name, field_type):
        from .types.base import BaseType
        from .types.serializable import Serializable

        assert isinstance(field_type, (BaseType, Serializable))
        self.name = name
        self.type = field_type

    def is_settable(self):
        return getattr(self.type, "fset", None) is not None



import uuid

import pytest

from schematics.common import *
from schematics.exceptions import *
from schematics.models import Model
from schematics.types import *
from schematics.types.compound import *
from schematics.types.serializable import Serializable


def test_id_or_uuid():

    id_field = UnionType((IntType, UUIDType))

    result = id_field.convert("ee5a16cb-0ee1-46bc-ac40-fb3018b1d29b")
    assert result == uuid.UUID("ee5a16cb-0ee1-46bc-ac40-fb3018b1d29b")
    assert type(id_field.to_primitive(result)) is str

    result = id_field.convert("99999")
    assert result == 99999
    assert type(result) is int
    assert type(id_field.to_primitive(result)) is int


def test_list_of_numbers():

    numbers = ListType(UnionType((IntType, FloatType)))
    result = numbers.convert(["2", "0.5", "123", "2.999"], None)
    assert result == [2, 0.5, 123, 2.999]
    assert [type(item) for item in result] == [int, float, int, float]
    assert [type(item) for item in numbers.to_primitive(result)] == [
        int,
        float,
        int,
        float,
    ]


def test_dict_or_list_of_ints():

    ints = UnionType((DictType, ListType), field=IntType)

    result = ints.convert(["1", "2", "3", "4"])
    assert result == [1, 2, 3, 4]
    assert ints.to_primitive(result) == [1, 2, 3, 4]

    result = ints.convert(dict(a=1, b=2, c=3, d=4))
    assert result == dict(a=1, b=2, c=3, d=4)
    assert [type(item) for item in result.values()] == [int, int, int, int]
    assert ints.to_primitive(result) == dict(a=1, b=2, c=3, d=4)


def test_custom_type():
    class ShoutType(StringType):
        def to_primitive(self, value, context):
            return StringType.convert(self, value) + "!!!"

    class QuestionType(StringType):
        def to_primitive(self, value, context):
            return StringType.convert(self, value) + "???"

    class FancyStringType(UnionType):
        types = (ShoutType, QuestionType)

        def resolve(self, value, context):
            if value.endswith("?"):
                return QuestionType
            elif value.endswith("!"):
                return ShoutType

    assert FancyStringType().to_primitive("Hello, world!") == "Hello, world!!!!"
    assert FancyStringType().to_primitive("Who's a good boy?") == "Who's a good boy????"

    def resolve(value, context):
        if value.endswith("?"):
            return QuestionType
        elif value.endswith("!"):
            return ShoutType

    assert (
        UnionType((ShoutType, QuestionType), resolver=resolve).to_primitive(
            "Hello, world!"
        )
        == "Hello, world!!!!"
    )

    with pytest.raises(ValidationError):
        FancyStringType(max_length=20).validate("What is the meaning of life?")


def test_option_collation():
    class DerivedStringType(StringType):
        pass

    field = UnionType(
        (IntType, DerivedStringType), required=True, max_value=100, max_length=50
    )
    assert field.required is True
    assert field._types[IntType].required is True
    assert field._types[DerivedStringType].required is True
    assert field._types[IntType].max_value == 100
    assert field._types[DerivedStringType].max_length == 50


def test_parameterize_validate():

    UnionType((IntType(min_value=1, strict=True), FloatType(min_value=0))).validate(
        0.00
    )

    UnionType((IntType(min_value=1), FloatType(min_value=0.5))).validate(0.75)
    with pytest.raises(ValidationError):
        UnionType((IntType(min_value=1), FloatType(min_value=0.5))).validate(0.00)


def test_invalid_args():

    with pytest.raises(TypeError):
        UnionType((IntType, dict))



import pytest

from schematics.common import *
from schematics.models import Model
from schematics.types import *
from schematics.types.compound import *
from schematics.types.serializable import serializable
from schematics.undefined import *
from schematics.util import *


def test_undefined():

    assert str(Undefined) == "Undefined"
    assert repr(Undefined) == "Undefined"

    assert Undefined == Undefined
    assert not Undefined != Undefined

    assert bool(Undefined) is False

    assert Undefined not in (None, True, False, 0, 1)

    with pytest.raises(TypeError):
        int(Undefined)

    with pytest.raises(TypeError):
        Undefined < 0
    with pytest.raises(TypeError):
        Undefined > 0
    with pytest.raises(TypeError):
        Undefined >= 0
    with pytest.raises(TypeError):
        Undefined >= 0

    with pytest.raises(TypeError):
        Undefined()

    with pytest.raises(TypeError):
        Undefined.x = 1
    with pytest.raises(AttributeError):
        Undefined.x

    assert UndefinedType() is Undefined

    with pytest.raises(TypeError):

        class DerivativeType(UndefinedType):
            pass

        DerivativeType()


def test_setdefault():
    class A:
        y = 2
        z = None

    class B(A):
        pass

    b = B()
    result = setdefault(b, "i", 9)
    assert result == b.i == 9
    with pytest.raises(AttributeError):
        B.i

    b = B()
    b.i = 1
    result = setdefault(b, "i", 9)
    assert result == b.i == 1

    b = B()
    b.i = None
    result = setdefault(b, "i", 9)
    assert result is b.i is None

    b = B()
    b.i = None
    result = setdefault(b, "i", 9, overwrite_none=True)
    assert result == b.i == 9
    assert B.y == 2

    b = B()
    result = setdefault(b, "y", 9)
    assert result == b.y == 9
    assert B.y == 2

    b = B()
    result = setdefault(b, "y", 9, overwrite_none=True)
    assert result == b.y == 9
    assert B.y == 2

    b = B()
    result = setdefault(b, "y", 9, search_mro=True)
    assert result == b.y == 2

    b = B()
    result = setdefault(b, "z", 9, search_mro=True)
    assert result is b.z is None

    b = B()
    result = setdefault(b, "z", 9, search_mro=True, overwrite_none=True)
    assert result == b.z == 9
    assert B.z is None


def test_constant():

    C = Constant("C", 99)
    assert C == 99
    assert C.name == "C"

    with pytest.raises(ValueError):
        C = Constant("C", "foo")
